from mgraph_ai_service_cache import package_name

SERVICE_NAME                             = package_name
FAST_API__TITLE                          = "MGraph-AI Service Cache"
FAST_API__DESCRIPTION                    = "Cache Service"
LAMBDA_DEPENDENCIES__FAST_API_SERVERLESS = ['osbot-fast-api-serverless==1.32.0'    ,
                                            'memory-fs==v0.37.0'                    ,
                                            'mgraph-ai-service-cache-client==v0.22.0']

CACHE_SERVICE__WEB_CONSOLE__PATH              = 'console'
CACHE_SERVICE__WEB_CONSOLE__ROUTE__START_PAGE = 'index'
CACHE_SERVICE__WEB_CONSOLE__MAJOR__VERSION    = "v0/v0.1"
CACHE_SERVICE__WEB_CONSOLE__LATEST__VERSION   = "v0.1.1"
ROUTES_PATHS__WEB_CONSOLE                     = [f'/{CACHE_SERVICE__WEB_CONSOLE__PATH}']


import hashlib
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash import Safe_Str__Cache_Hash
from osbot_utils.utils.Zip                                                               import zip_bytes__files


def zip_bytes__content_hash(zip_bytes: bytes, hash_length: int) -> Safe_Str__Cache_Hash:     # Calculate hash based on ZIP content, not raw bytes. This ensures identical content produces the same hash regardless of  creation time or compression settings.

    files_dict  = zip_bytes__files(zip_bytes)                                       # Use existing zip_bytes__files to get all files and their content
    sorted_files = sorted(files_dict.items())                                       # Sort files by name for deterministic ordering
    hasher = hashlib.sha256()                                                       # Create hash from file paths and contents

    for file_path, file_content in sorted_files:                                    # Hash the file path
        hasher.update(file_path.encode('utf-8'))
        hasher.update(b'\x00'                  )                                    # Null separator between path and content
        hasher.update(file_content             )                                    # Hash the file content
        hasher.update(b'\x00'                  )                                    # Null separator between entries

    full_hash   = hasher.hexdigest()                                                # Get the full hash and truncate to configured length

    return Safe_Str__Cache_Hash(full_hash[:hash_length])


from typing                                                                              import Dict
from osbot_fast_api.api.routes.Fast_API__Routes                                          import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix               import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag                  import Safe_Str__Fast_API__Route__Tag
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id          import Safe_Str__Id
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash import Safe_Str__Cache_Hash
from mgraph_ai_service_cache.service.cache.Cache__Service                                import Cache__Service

TAG__ROUTES_EXISTS                  = 'exists'
PREFIX__ROUTES_EXISTS               = '/{namespace}'
BASE_PATH__ROUTES_EXISTS            = f'{PREFIX__ROUTES_EXISTS}/{TAG__ROUTES_EXISTS}/'
ROUTES_PATHS__EXISTS                = [ BASE_PATH__ROUTES_EXISTS + 'hash/{cache_hash}']


class Routes__File__Exists(Fast_API__Routes):
    tag           : Safe_Str__Fast_API__Route__Tag     = TAG__ROUTES_EXISTS
    prefix        : Safe_Str__Fast_API__Route__Prefix  =  PREFIX__ROUTES_EXISTS
    cache_service : Cache__Service


    # todo: add exists__hash__cache_id      # where we can resolve if a particular cache_id exists (without needing to use the cache

    # todo: refactor code below with the version above (which is using Schema__Cache__Exists__Response)
    def exists__hash__cache_hash(self, cache_hash  : Safe_Str__Cache_Hash                                   ,  # Check if hash exists
                                       namespace   : Safe_Str__Id = None
                                  ) -> Dict[str, bool]:
        namespace = namespace or Safe_Str__Id("default")                                                             # todo: refactor to use static var for default namespace
        handler   = self.cache_service.get_or_create_handler(namespace)
        file_id   = Safe_Str__Id(cache_hash)                                                                         # cast hash to File_Id (which is what file__json will use)

        with handler.fs__refs_hash.file__json__single(file_id) as ref_fs:
            exists = ref_fs.exists()

        return {"exists"     : exists         ,
                "cache_hash" : str(cache_hash),
                "namespace"  : str(namespace )}

    def setup_routes(self):
        self.add_route_get(self.exists__hash__cache_hash)


from typing                                                                     import Dict, Any
from osbot_fast_api.api.routes.Fast_API__Routes                                 import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix      import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag         import Safe_Str__Fast_API__Route__Tag
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id              import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id import Safe_Str__Id
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API              import FAST_API__PARAM__NAMESPACE
from mgraph_ai_service_cache.service.cache.Cache__Service                       import Cache__Service

TAG__ROUTES_DELETE                  = 'delete'
PREFIX__ROUTES_DELETE               = '/{namespace}'
BASE_PATH__ROUTES_DELETE            = f'{PREFIX__ROUTES_DELETE}/{TAG__ROUTES_DELETE}/'
ROUTES_PATHS__DELETE                = [ BASE_PATH__ROUTES_DELETE + '{cache_id}']


class Routes__File__Delete(Fast_API__Routes):
    tag           : Safe_Str__Fast_API__Route__Tag     = TAG__ROUTES_DELETE
    prefix        : Safe_Str__Fast_API__Route__Prefix  =  PREFIX__ROUTES_DELETE
    cache_service : Cache__Service

    # todo: this should a variation of the Schema__Cache__Delete__Success (which needed refactoring)
    def delete__cache_id(self, cache_id  : Cache_Id    ,
                               namespace : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                          ) -> Dict[str, Any]:
        return self.cache_service.delete_by_id(cache_id, namespace)


    def setup_routes(self):
        self.add_route_delete(self.delete__cache_id)


from typing                                                                     import Dict, Any
from osbot_fast_api.api.routes.Fast_API__Routes                                 import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix      import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag         import Safe_Str__Fast_API__Route__Tag
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id import Safe_Str__Id
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API              import FAST_API__PARAM__NAMESPACE
from mgraph_ai_service_cache.service.cache.Cache__Service                       import Cache__Service

TAG__ROUTES_NAMESPACE                  = 'namespace'
PREFIX__ROUTES_NAMESPACE               = '/{namespace}'
ROUTES_PATHS__NAMESPACE                = [ PREFIX__ROUTES_NAMESPACE + '/file-hashes',
                                           PREFIX__ROUTES_NAMESPACE + '/file-ids'   ,
                                           PREFIX__ROUTES_NAMESPACE + '/stats'      ]


class Routes__Namespace(Fast_API__Routes):
    tag           : Safe_Str__Fast_API__Route__Tag     = TAG__ROUTES_NAMESPACE
    prefix        : Safe_Str__Fast_API__Route__Prefix  =  PREFIX__ROUTES_NAMESPACE
    cache_service : Cache__Service

    # todo: BUG: rename form file_hashes to cache_hashes
    def file_hashes(self, namespace: Safe_Str__Id = FAST_API__PARAM__NAMESPACE):
        return self.cache_service.get_namespace__file_hashes(namespace=namespace)

    # todo: BUG: rename form file_hashes to cache_hashes
    def file_ids(self, namespace: Safe_Str__Id = FAST_API__PARAM__NAMESPACE):
        return self.cache_service.get_namespace__file_ids(namespace=namespace)

    def stats(self, namespace: Safe_Str__Id = FAST_API__PARAM__NAMESPACE) -> Dict[str, Any]:       # Get cache statistics
        namespace = namespace or Safe_Str__Id("default")

        try:
            # Get file counts using shared method
            counts_data = self.cache_service.get_namespace__file_counts(namespace)
            handler     = counts_data['handler']
            # Build stats response
            stats = { "namespace": str(namespace)         ,
                      "ttl_hours": handler.cache_ttl_hours,
                      **counts_data['file_counts'        ]}  # Spread all the file counts

            return stats
        except Exception as e:
            return {"error": str(e), "namespace": str(namespace)}

    def setup_routes(self):
        self.add_route_get(self.file_hashes)
        self.add_route_get(self.file_ids   )
        self.add_route_get(self.stats      )


from memory_fs.storage_fs.Storage_FS                                                     import Storage_FS
from memory_fs.storage_fs.providers.Storage_FS__Local_Disk                               import Storage_FS__Local_Disk
from memory_fs.storage_fs.providers.Storage_FS__Memory                                   import Storage_FS__Memory
from memory_fs.storage_fs.providers.Storage_FS__Sqlite                                   import Storage_FS__Sqlite
from memory_fs.storage_fs.providers.Storage_FS__Zip                                      import Storage_FS__Zip
from mgraph_ai_service_cache_client.schemas.consts.const__Storage                        import ENV_VAR__CACHE__SERVICE__LOCAL_DISK_PATH, ENV_VAR__CACHE__SERVICE__SQLITE_PATH, ENV_VAR__CACHE__SERVICE__ZIP_PATH, ENV_VAR__CACHE__SERVICE__STORAGE_MODE
from mgraph_ai_service_cache.service.storage.Storage_FS__S3                              import Storage_FS__S3
from mgraph_ai_service_cache.utils.for_osbot_utils.Env                                   import get_env_enum
from osbot_utils.type_safe.Type_Safe                                                     import Type_Safe
from osbot_utils.type_safe.primitives.core.Safe_UInt                                     import Safe_UInt
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id          import Safe_Str__Id
from osbot_utils.utils.Env                                                               import get_env
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Storage_Mode        import Enum__Cache__Storage_Mode
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API                       import (ENV_VAR__CACHE__SERVICE__BUCKET_NAME,
                                                                                                 ENV_VAR__CACHE__SERVICE__DEFAULT_TTL_HOURS,
                                                                                                 DEFAULT__CACHE__SERVICE__BUCKET_NAME,
                                                                                                 DEFAULT__CACHE__SERVICE__DEFAULT_TTL_HOURS)


# todo: refactor all the parms below to an Schema__Cache__Config
class Cache__Config(Type_Safe):                                                             # Configuration for cache service
    storage_mode      : Enum__Cache__Storage_Mode     = None                                # Storage backend mode
    default_bucket    : Safe_Str__Id                  = None                                # S3 bucket (for S3 mode)
    default_ttl_hours : Safe_UInt                     = None                                # TTL for cache entries
    local_disk_path   : str                           = None                                # Path for local disk storage
    sqlite_path       : str                           = None                                # Path for SQLite storage
    zip_path          : str                           = None                                # Path for ZIP storage

    # todo: see if we can move this __init__ actions to a setup() class since it is never good to have any changes done on __init__
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self.storage_mode is None:                                                       # Auto-configure if values not provided
            self.storage_mode = self.determine_storage_mode()

        self.configure_for_storage_mode()                                                   # Set defaults based on storage mode

    def determine_storage_mode(self) -> Enum__Cache__Storage_Mode:                                          # Auto-detect best storage mode
        explicit_mode = get_env_enum(ENV_VAR__CACHE__SERVICE__STORAGE_MODE, Enum__Cache__Storage_Mode)      # Check for explicit mode configuration
        if explicit_mode:
            return explicit_mode

        if self.has_aws_credentials():                                                                      # Check if AWS credentials are available for S3
            return Enum__Cache__Storage_Mode.S3

        return Enum__Cache__Storage_Mode.MEMORY                                                             # Default to memory mode

    # todo: refactor this with a check from osbot_aws
    def has_aws_credentials(self) -> bool:                                                  # Check if AWS credentials are available
        # Check for AWS credentials in various forms
        if get_env('AWS_ACCESS_KEY_ID') and get_env('AWS_SECRET_ACCESS_KEY'):
            return True
        if get_env('AWS_PROFILE'):
            return True
        if get_env('AWS_REGION') and get_env('AWS_LAMBDA_FUNCTION_NAME'):                   # In Lambda
            return True
        return False

    def configure_for_storage_mode(self):                                                   # Set configuration based on storage mode
        if self.default_ttl_hours is None:                                                  # Configure TTL (applies to all modes)
            self.default_ttl_hours = get_env(ENV_VAR__CACHE__SERVICE__DEFAULT_TTL_HOURS,
                                             DEFAULT__CACHE__SERVICE__DEFAULT_TTL_HOURS)

        if self.storage_mode == Enum__Cache__Storage_Mode.S3:                               # Mode-specific configuration
            if self.default_bucket is None:
                self.default_bucket = get_env(ENV_VAR__CACHE__SERVICE__BUCKET_NAME,
                                              DEFAULT__CACHE__SERVICE__BUCKET_NAME)

        elif self.storage_mode == Enum__Cache__Storage_Mode.LOCAL_DISK:
            if self.local_disk_path is None:
                self.local_disk_path = get_env(ENV_VAR__CACHE__SERVICE__LOCAL_DISK_PATH, '/tmp/cache')              # todo: refactor this value into a static config variable

        elif self.storage_mode == Enum__Cache__Storage_Mode.SQLITE:
            if self.sqlite_path is None:
                self.sqlite_path = get_env(ENV_VAR__CACHE__SERVICE__SQLITE_PATH, ':memory:')                        # todo: refactor this value into a static config variable

        elif self.storage_mode == Enum__Cache__Storage_Mode.ZIP:
            if self.zip_path is None:
                self.zip_path = get_env(ENV_VAR__CACHE__SERVICE__ZIP_PATH, '/tmp/cache.zip')                        # todo: refactor this value into a static config variable

    def create_storage_backend(self) -> Storage_FS:                                                                 # Create the appropriate storage backend
        if self.storage_mode == Enum__Cache__Storage_Mode.MEMORY:
            return Storage_FS__Memory()

        elif self.storage_mode == Enum__Cache__Storage_Mode.S3:
            if not self.default_bucket:
                raise ValueError("S3 bucket name required for S3 storage mode")
            return Storage_FS__S3(s3_bucket=self.default_bucket).setup()

        elif self.storage_mode == Enum__Cache__Storage_Mode.LOCAL_DISK:
            return Storage_FS__Local_Disk(root_path=self.local_disk_path)

        elif self.storage_mode == Enum__Cache__Storage_Mode.SQLITE:
            in_memory = self.sqlite_path == ':memory:'
            return Storage_FS__Sqlite(db_path=self.sqlite_path, in_memory=in_memory).setup()

        elif self.storage_mode == Enum__Cache__Storage_Mode.ZIP:
            return Storage_FS__Zip(zip_path=self.zip_path, in_memory=False).setup()

    # todo see if can reuse Schema__Cache__Config (or if we need a new schema class)
    def get_storage_info(self) -> dict:                                                     # Get information about current configuration
        info = { 'storage_mode' : self.storage_mode.value   ,
                 'ttl_hours'    : self.default_ttl_hours    }                               # todo: refactor to Type_Safe class

        if self.storage_mode == Enum__Cache__Storage_Mode.S3:
            info['s3_bucket'] = self.default_bucket
        elif self.storage_mode == Enum__Cache__Storage_Mode.LOCAL_DISK:
            info['local_disk_path'] = self.local_disk_path
        elif self.storage_mode == Enum__Cache__Storage_Mode.SQLITE:
            info['sqlite_path'] = self.sqlite_path
        elif self.storage_mode == Enum__Cache__Storage_Mode.ZIP:
            info['zip_path'] = self.zip_path

        return info                                                                          # todo: refactor to Type_Safe class


from osbot_utils.type_safe.Type_Safe                                            import Type_Safe
from osbot_utils.type_safe.primitives.domains.common.safe_str.Safe_Str__Version import Safe_Str__Version
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id import Safe_Str__Id
from mgraph_ai_service_cache.config                                             import SERVICE_NAME
from mgraph_ai_service_cache.service.info.schemas.Enum__Service_Environment     import Enum__Service_Environment
from mgraph_ai_service_cache.service.info.schemas.Enum__Service_Status          import Enum__Service_Status
from mgraph_ai_service_cache.utils.Version                                      import version__mgraph_ai_service_cache

class Schema__Service__Status(Type_Safe):
    name        : Safe_Str__Id              = Safe_Str__Id(SERVICE_NAME)
    version     : Safe_Str__Version         = version__mgraph_ai_service_cache
    status      : Enum__Service_Status      = Enum__Service_Status.operational
    environment : Enum__Service_Environment = Enum__Service_Environment.local


from typing                                                                               import Optional, Dict, Any
from memory_fs.schemas.Schema__Memory_FS__File__Metadata                                  import Schema__Memory_FS__File__Metadata
from memory_fs.schemas.Schema__Memory_FS__File__Config                                    import Schema__Memory_FS__File__Config
from osbot_utils.type_safe.Type_Safe                                                      import Type_Safe
from osbot_utils.type_safe.primitives.core.Safe_UInt                                      import Safe_UInt
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                        import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_int.Timestamp_Now          import Timestamp_Now
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id           import Safe_Str__Id
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                            import type_safe
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash  import Safe_Str__Cache_Hash
from mgraph_ai_service_cache_client.schemas.cache.file.Schema__Cache__File__Refs          import Schema__Cache__File__Refs
from mgraph_ai_service_cache_client.schemas.cache.Schema__Cache__Metadata                 import Schema__Cache__Metadata
from mgraph_ai_service_cache_client.schemas.cache.Schema__Cache__Retrieve__Success        import Schema__Cache__Retrieve__Success
from mgraph_ai_service_cache_client.schemas.cache.consts__Cache_Service                   import DEFAULT_CACHE__NAMESPACE
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Data_Type            import Enum__Cache__Data_Type
from mgraph_ai_service_cache_client.schemas.errors.Schema__Cache__Error__Gone             import Schema__Cache__Error__Gone
from mgraph_ai_service_cache_client.schemas.errors.Schema__Cache__Error__Not_Found        import Schema__Cache__Error__Not_Found
from mgraph_ai_service_cache.service.cache.Cache__Service                                 import Cache__Service

# todo: rename to Cache__Service__Retrieve

class Cache__Service__Retrieve(Type_Safe):                                            # Service layer for cache retrieval operations
    cache_service : Cache__Service                                                    # Underlying cache service

    @type_safe
    def check_exists(self, cache_hash : Safe_Str__Cache_Hash,
                           namespace  : Safe_Str__Id        = DEFAULT_CACHE__NAMESPACE
                     ) -> bool:                                                       # Check if cache entry exists

        handler   = self.cache_service.get_or_create_handler(namespace)

        with handler.fs__refs_hash.file__json__single(Safe_Str__Id(str(cache_hash))) as ref_fs:
            return ref_fs.exists()

    @type_safe
    def retrieve_by_hash(self, cache_hash : Safe_Str__Cache_Hash,
                               namespace  : Safe_Str__Id        = DEFAULT_CACHE__NAMESPACE
                         ) -> Optional[Schema__Cache__Retrieve__Success]:               # Retrieve entry by content hash

        result = self.cache_service.retrieve_by_hash(cache_hash, namespace)             # Use cache service to get the data
        
        if result is None:
            return None                                                                 # Caller decides how to handle not found

        metadata = self._build_metadata(result)                                         # Convert to Type_Safe response schema
        
        return Schema__Cache__Retrieve__Success(data      = result.get("data")     ,
                                                metadata  = metadata               ,
                                                data_type = self._determine_data_type(result))

    @type_safe
    def retrieve_by_hash__metadata(self,                                                                # Retrieve metadata by content hash
                                   cache_hash : Safe_Str__Cache_Hash,
                                   namespace  : Safe_Str__Id        = DEFAULT_CACHE__NAMESPACE
                              ) -> Schema__Cache__Metadata:
        with self.cache_service as _:
            result   = _.retrieve_by_hash(cache_hash, namespace)                           # Use cache service to get the data
            if result is None:
                return None
            metadata = self._build_metadata(result)                                         # Convert to Type_Safe response schema
            return metadata

    # todo: change return to type_safe value
    @type_safe
    def retrieve_by_hash__refs_hash(self,                                                                # Retrieve metadata by content hash
                                   cache_hash : Safe_Str__Cache_Hash,
                                   namespace  : Safe_Str__Id        = DEFAULT_CACHE__NAMESPACE
                              ) -> Dict:
        with self.cache_service as _:
            refs_hash   = _.retrieve_by_hash__refs_hash(cache_hash, namespace)                           # Use cache service to get the data
            return refs_hash
    
    @type_safe
    def retrieve_by_id(self, cache_id  : Cache_Id,
                             namespace : Safe_Str__Id = DEFAULT_CACHE__NAMESPACE
                        ) -> Optional[Schema__Cache__Retrieve__Success]:                # Retrieve entry by cache ID

        result = self.cache_service.retrieve_by_id(cache_id, namespace)                 # Use cache service to get the data
        
        if result is None:
            return None

        metadata = self._build_metadata(result)                                         # Convert to Type_Safe response
        
        return Schema__Cache__Retrieve__Success(data      = result.get("data")     ,
                                                metadata  = metadata               ,
                                                data_type = self._determine_data_type(result))

    @type_safe
    def retrieve_by_id__config(self, cache_id  : Cache_Id,
                             namespace : Safe_Str__Id = None
                        ) -> Schema__Memory_FS__File__Config:                                   # Retrieve entry by cache ID

        return self.cache_service.retrieve_by_id__config(cache_id, namespace)                   #

    @type_safe
    def retrieve_by_id__metadata(self, cache_id  : Cache_Id,
                                       namespace : Safe_Str__Id = DEFAULT_CACHE__NAMESPACE
                                 ) -> Schema__Memory_FS__File__Metadata:                         # Get contents of the by-id metadata file
        return self.cache_service.retrieve_by_id__metadata(cache_id, namespace)                 #

    @type_safe
    def retrieve_by_id__refs(self, cache_id  : Cache_Id,
                                   namespace : Safe_Str__Id = DEFAULT_CACHE__NAMESPACE
                             ) -> Schema__Cache__File__Refs:                                    # Get contents of the by-id refs file
        return self.cache_service.retrieve_by_id__refs(cache_id, namespace)                     #


    def get_entry_details__all(self, cache_id  : Cache_Id,
                                     namespace : Safe_Str__Id = DEFAULT_CACHE__NAMESPACE
                                ) -> Optional[Dict[str, Any]]:                                # Get detailed information about cache entry
        details  = self.retrieve_by_id__refs(cache_id=cache_id, namespace=namespace)
        if not details:
            return None

        all_details   = {}
        paths__content = details.file_paths.content_files                                     # capture this path, since we don't want to show it (i.e. don't add the file's content)
        storage_fs    = self.cache_service.storage_fs()
        for file_type, file_paths in details.all_paths.json().items():
            for file_path in file_paths:
                if file_path not in paths__content:
                    file_contents          = storage_fs.file__json(file_path)          # all these files are json files
                    all_details[file_path] = file_contents
        return dict(by_id   =  details.json()   ,
                    details = all_details)

    @type_safe
    def get_not_found_error(self, resource_id : Safe_Str__Id         = None,
                                  cache_hash  : Safe_Str__Cache_Hash = None,
                                  cache_id    : Cache_Id          = None,
                                  namespace   : Safe_Str__Id         = None
                            ) -> Schema__Cache__Error__Not_Found:                                                          # Build not found error response
        return Schema__Cache__Error__Not_Found(error_type    = "NOT_FOUND"                                           ,     # todo: refactor error_type to Enum
                                               message       = "The requested cache entry was not found"             ,
                                               resource_type = "cache_entry"                                         ,
                                               resource_id   = resource_id                                           ,
                                               cache_hash    = cache_hash                                            ,
                                               cache_id      = cache_id                                              ,
                                               namespace     = namespace                                              )
    
    @type_safe
    def get_expired_error(self, cache_id   : Cache_Id   = None,
                                expired_at : Timestamp_Now = None,
                                ttl_hours  : Safe_UInt     = None,
                                namespace  : Safe_Str__Id  = None
                          ) -> Schema__Cache__Error__Gone:                                                                  # Build expired error response
        return Schema__Cache__Error__Gone(error_type = "EXPIRED"                                                ,           # todo: refactor error_type to Enum
                                          message    = "The cache entry has expired and is no longer available" ,
                                          cache_id   = cache_id                                                 ,
                                          expired_at = expired_at                                               ,
                                          ttl_hours  = ttl_hours                                                ,
                                          namespace  = namespace                                                 )
    
    def _build_metadata(self, cache_result: Dict[str, Any]) -> Schema__Cache__Metadata:                                     # Build metadata from cache result
        metadata_raw = cache_result.get("metadata", {})

        # todo: this should be able to just be Schema__Cache__Metadata.from_json(metadata_raw),
        #       in fact metadata_raw should be Type_Safe
        return Schema__Cache__Metadata(cache_id         = metadata_raw.get("cache_id")         ,
                                       cache_hash       = metadata_raw.get("cache_hash")       ,
                                       cache_key        = metadata_raw.get("cache_key")        ,
                                       file_id          = metadata_raw.get("file_id")          ,
                                       namespace        = metadata_raw.get("namespace")        ,
                                       strategy         = metadata_raw.get("strategy")         ,
                                       stored_at        = metadata_raw.get("stored_at")        ,
                                       file_type        = metadata_raw.get("file_type", "json"),
                                       content_encoding = metadata_raw.get("content_encoding") ,
                                       content_size     = metadata_raw.get("content__size", 0)  )

    # todo: we should just be able to use Enum__Cache__Data_Type from cache_result.get("data_type") and let Type_Safe handle the rest
    def _determine_data_type(self, cache_result: Dict[str, Any]) -> Enum__Cache__Data_Type:  # Determine data type from result
        data_type_str = cache_result.get("data_type", "string")
        
        if data_type_str == "binary":
            return Enum__Cache__Data_Type.BINARY
        elif data_type_str == "json":
            return Enum__Cache__Data_Type.JSON
        else:
            return Enum__Cache__Data_Type.STRING
    
    def _is_expired(self, id_ref: Dict[str, Any]) -> bool:                           # Check if cache entry has expired
        # TODO: Implement TTL expiration logic
        # For now, always return False (nothing expires)
        return False


from osbot_fast_api.api.routes.Fast_API__Routes                                 import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix      import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag         import Safe_Str__Fast_API__Route__Tag
from osbot_utils.decorators.methods.cache_on_self                               import cache_on_self
from mgraph_ai_service_cache.service.cache.Cache__Service                       import Cache__Service

TAG__ROUTES_NAMESPACES      = 'namespaces'
PREFIX__ROUTES_NAMESPACES   = '/namespaces'
ROUTES_PATHS__NAMESPACES    = [ PREFIX__ROUTES_NAMESPACES + '/list']


class Routes__Namespaces(Fast_API__Routes):
    tag           : Safe_Str__Fast_API__Route__Tag     = TAG__ROUTES_NAMESPACES
    prefix        : Safe_Str__Fast_API__Route__Prefix  = PREFIX__ROUTES_NAMESPACES
    cache_service : Cache__Service

    @cache_on_self
    def storage_fs(self):
        return self.cache_service.storage_fs()

    # todo, move this to a namespaces service
    def list(self):
        namespaces = self.storage_fs().folder__folders(parent_folder='/', return_full_path=False)       # the list of namespaces is the list of root folders
        return namespaces

    def setup_routes(self):
        self.add_route_get(self.list)


from fastapi                                                                                    import HTTPException, Request, Body
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Json__Field_Path   import Safe_Str__Json__Field_Path
from mgraph_ai_service_cache.service.cache.Cache__Service                                       import Cache__Service
from osbot_fast_api.api.decorators.route_path                                                   import route_path
from osbot_fast_api.api.routes.Fast_API__Routes                                                 import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix                      import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag                         import Safe_Str__Fast_API__Route__Tag
from osbot_utils.decorators.methods.cache_on_self                                               import cache_on_self
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path               import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id                 import Safe_Str__Id
from mgraph_ai_service_cache_client.schemas.cache.consts__Cache_Service                         import DEFAULT_CACHE__STORE__STRATEGY
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Store__Strategy            import Enum__Cache__Store__Strategy
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API                              import FAST_API__PARAM__NAMESPACE
from mgraph_ai_service_cache.service.cache.store.Cache__Service__Store                          import Cache__Service__Store
from mgraph_ai_service_cache_client.schemas.cache.Schema__Cache__Store__Response                import Schema__Cache__Store__Response

TAG__ROUTES_STORE                  = 'store'
PREFIX__ROUTES_STORE               = '/{namespace}/{strategy}'
ROUTES_PATHS__STORE                = [ f'{PREFIX__ROUTES_STORE}/{TAG__ROUTES_STORE}/' + 'string'                  ,
                                       f'{PREFIX__ROUTES_STORE}/{TAG__ROUTES_STORE}/' + 'string/{cache_key:path}' ,
                                       f'{PREFIX__ROUTES_STORE}/{TAG__ROUTES_STORE}/' + 'json'                    ,
                                       f'{PREFIX__ROUTES_STORE}/{TAG__ROUTES_STORE}/' + 'json/{cache_key:path}'   ,
                                       f'{PREFIX__ROUTES_STORE}/{TAG__ROUTES_STORE}/' + 'binary'                  ,
                                       f'{PREFIX__ROUTES_STORE}/{TAG__ROUTES_STORE}/' + 'binary/{cache_key:path}' ]


class Routes__File__Store(Fast_API__Routes):                                                                  # FastAPI routes for cache store operations
    tag           : Safe_Str__Fast_API__Route__Tag    = TAG__ROUTES_STORE
    prefix        : Safe_Str__Fast_API__Route__Prefix = PREFIX__ROUTES_STORE
    cache_service : Cache__Service                                                                     # get cache_service via Dependency Injection

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @cache_on_self
    def store_service(self):                                                                            # Service layer for business logic
        return Cache__Service__Store(cache_service=self.cache_service)                                  # create Cache__Service__Store object (once, using the shared Cache_Service)


    def store__string(self, data      : str = Body(...),
                            strategy  : Enum__Cache__Store__Strategy = DEFAULT_CACHE__STORE__STRATEGY,
                            namespace : Safe_Str__Id                 = FAST_API__PARAM__NAMESPACE
                       ) -> Schema__Cache__Store__Response:                                                         # Store string data

        if not data:                                                                                                # Validate input
            error = self.store_service().get_invalid_input_error( field_name    = "data"                       ,      # todo: see if we should support this case where the files are empty
                                                                  expected_type = "non-empty string"           ,
                                                                  message       = "String data cannot be empty")
            raise HTTPException(status_code=400, detail=error.json())

        # Use service layer
        result = self.store_service().store_string(data      = data       ,
                                                 strategy  = strategy   ,
                                                 namespace = namespace  )

        if result is None:
            raise HTTPException(status_code=500, detail="Failed to store data")

        return result

    @route_path("/store/string/{cache_key:path}")
    def store__string__cache_key(self, data       : str                          = Body(...)                        ,
                                       namespace  : Safe_Str__Id                 = FAST_API__PARAM__NAMESPACE         ,
                                       strategy   : Enum__Cache__Store__Strategy = DEFAULT_CACHE__STORE__STRATEGY   ,
                                       cache_key  : Safe_Str__File__Path         = None,
                                       file_id    : Safe_Str__Id                 = None
                                  ) -> Schema__Cache__Store__Response:                                  # Store string with semantic key

        if not data:                                                                                    # Validate input
            error = self.store_service().get_invalid_input_error(field_name    = "data"                       ,
                                                                 expected_type = "non-empty string"           ,
                                                                 message       = "String data cannot be empty")
            raise HTTPException(status_code=400, detail=error.json())

        if not cache_key:
            error = self.store_service().get_invalid_input_error(field_name    = "cache_key"                              ,
                                                                 expected_type = "valid file path"                        ,
                                                                 message       = "Cache key is required for this endpoint")
            raise HTTPException(status_code=400, detail=error.json())

        result = self.store_service().store_string(data      = data       ,                               # Use service layer
                                                   strategy  = strategy   ,                               # todo: we should we using a Type_Safe class for these params
                                                   namespace = namespace  ,
                                                   cache_key = cache_key  ,
                                                   file_id   = file_id    )

        if result is None:
            raise HTTPException(status_code=500, detail="Failed to store data")

        return result

    def store__json(self, data     : dict,                                                                          # JSON can be empty object, that's valid
                          strategy  : Enum__Cache__Store__Strategy = DEFAULT_CACHE__STORE__STRATEGY,
                          namespace : Safe_Str__Id                 = FAST_API__PARAM__NAMESPACE
                     ) -> Schema__Cache__Store__Response:                                                           # Store JSON data

        result = self.store_service().store_json(data      = data     ,                                               # Use service layer
                                               strategy  = strategy ,                                               # todo: we should we using a Type_Safe class for these params
                                               namespace = namespace)

        if result is None:
            raise HTTPException(status_code=500, detail="Failed to store data")

        return result

    @route_path("/store/json/{cache_key:path}")
    def store__json__cache_key(self, data            : dict                         = Body(...)                     ,
                                     namespace       : Safe_Str__Id                 = FAST_API__PARAM__NAMESPACE    ,
                                     strategy        : Enum__Cache__Store__Strategy = DEFAULT_CACHE__STORE__STRATEGY,
                                     cache_key       : Safe_Str__File__Path         = None                          ,
                                     file_id         : Safe_Str__Id                 = None                          ,
                                     json_field_path : Safe_Str__Json__Field_Path   = None
                                ) -> Schema__Cache__Store__Response:                                                # Store JSON with semantic key

        if not cache_key:                                                                                           # todo: check this path, since I think this path can't be reached (due to how FastAPI handles routes)
            error = self.store_service().get_invalid_input_error(field_name    = "cache_key",
                                                               expected_type = "valid file path",
                                                               message       = "Cache key is required for this endpoint")
            raise HTTPException(status_code=400, detail=error.json())

        result = self.store_service().store_json(data            = data     ,                                               # Use service layer
                                                 strategy        = strategy ,                                               # todo: we should we using a Type_Safe class for these params
                                                 namespace       = namespace,
                                                 cache_key       = cache_key,
                                                 json_field_path = json_field_path,
                                                 file_id         = file_id  )

        if result is None:
            raise HTTPException(status_code=500, detail="Failed to store data")

        return result

    def store__binary(self, request  : Request,
                            body     : bytes                        = Body(..., media_type="application/octet-stream"),
                            strategy : Enum__Cache__Store__Strategy = DEFAULT_CACHE__STORE__STRATEGY,
                            namespace: Safe_Str__Id                 = FAST_API__PARAM__NAMESPACE
                       ) -> Schema__Cache__Store__Response:                                                         # Store binary data

        if not body:                                                                                                # Validate input
            error = self.store_service().get_invalid_input_error(field_name    = "body"                       ,
                                                               expected_type = "non-empty binary data"      ,
                                                               message       = "Binary data cannot be empty")
            raise HTTPException(status_code=400, detail=error.json())

        # todo: we need to refactor this gzip/encoding support into separate endpoint
        content_encoding = request.headers.get('content-encoding')                                                  # Check for compression

        result = self.store_service().store_binary(data             = body            ,                               # Use service layer
                                                 strategy         = strategy        ,                               # todo: we should we using a Type_Safe class for these params
                                                 namespace        = namespace       ,
                                                 content_encoding = content_encoding)

        if result is None:
            raise HTTPException(status_code=500, detail="Failed to store data")

        return result

    @route_path("/store/binary/{cache_key:path}")
    def store__binary__cache_key(self, body       : bytes                        = Body(..., media_type="application/octet-stream"),
                                       namespace  : Safe_Str__Id                 = FAST_API__PARAM__NAMESPACE      ,
                                       strategy   : Enum__Cache__Store__Strategy = DEFAULT_CACHE__STORE__STRATEGY,
                                       cache_key  : Safe_Str__File__Path         = None,
                                       file_id    : Safe_Str__Id                 = None,
                                       request    : Request                      = None
                                  ) -> Schema__Cache__Store__Response:                                              # Store binary with semantic key

        if not body:                                                                                                # Validate input
            error = self.store_service().get_invalid_input_error(field_name    = "body"                       ,       # todo: see if we should support this case where the files are empty
                                                               expected_type = "non-empty binary data"      ,
                                                               message       = "Binary data cannot be empty")
            raise HTTPException(status_code=400, detail=error.json())

        if not cache_key:
            error = self.store_service().get_invalid_input_error(field_name    = "cache_key"                              ,
                                                               expected_type = "valid file path"                        ,
                                                               message       = "Cache key is required for this endpoint")
            raise HTTPException(status_code=400, detail=error.json())

        # todo: we need to refactor this gzip/encoding support into separate endpoint
        content_encoding = request.headers.get('content-encoding') if request else None                             # Check for compression

        result = self.store_service().store_binary(data             = body            ,                               # Use service layer
                                                   strategy         = strategy        ,                               # todo: we should we using a Type_Safe class for these params
                                                   namespace        = namespace       ,
                                                   cache_key        = cache_key       ,
                                                   file_id          = file_id         ,
                                                   content_encoding = content_encoding)

        if result is None:
            raise HTTPException(status_code=500, detail="Failed to store data")

        return result

    def setup_routes(self):                                                             # Configure all routes
        self.add_route_post(self.store__string              )                           # String endpoints
        self.add_route_post(self.store__string__cache_key   )

        self.add_route_post(self.store__json                )
        self.add_route_post(self.store__json__cache_key     )                           # JSON endpoints

        self.add_route_post(self.store__binary              )                           # Binary endpoints
        self.add_route_post(self.store__binary__cache_key   )


from osbot_fast_api.api.routes.Fast_API__Routes         import Fast_API__Routes
from mgraph_ai_service_cache.service.info.Service_Info  import Service_Info

TAG__ROUTES_INFO                  = 'info'
ROUTES_PATHS__INFO                = [ f'/{TAG__ROUTES_INFO}/health'  ,
                                      f'/{TAG__ROUTES_INFO}/server'  ,
                                      f'/{TAG__ROUTES_INFO}/status'  ,
                                      f'/{TAG__ROUTES_INFO}/versions']
ROUTES_INFO__HEALTH__RETURN_VALUE = {'status': 'ok'}

class Routes__Info(Fast_API__Routes):
    tag         : str          = 'info'
    service_info: Service_Info

    def health(self):
        return ROUTES_INFO__HEALTH__RETURN_VALUE

    def server(self):                                             # Get service versions
        return self.service_info.server_info()

    def status(self):                                               # Get service status information
        return self.service_info.service_info()

    def versions(self):                                             # Get service versions
        return self.service_info.versions()


    def setup_routes(self):
        self.add_route_get(self.health  )
        self.add_route_get(self.server  )
        self.add_route_get(self.status  )
        self.add_route_get(self.versions)


from typing                                                                                     import Optional
from osbot_utils.decorators.methods.cache_on_self                                               import cache_on_self
from osbot_utils.type_safe.Type_Safe                                                            import Type_Safe
from osbot_utils.type_safe.primitives.domains.identifiers.Random_Guid                           import Random_Guid
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                                  import type_safe
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                              import Cache_Id
from osbot_utils.utils.Zip                                                                      import (zip_bytes__file_list, zip_bytes__file,
                                                                                                        zip_bytes__add_file, zip_bytes__remove_file,
                                                                                                        zip_bytes__replace_file)
from mgraph_ai_service_cache_client.schemas.cache.zip.enums.Enum__Cache__Zip__Operation         import Enum__Cache__Zip__Operation
from mgraph_ai_service_cache.service.cache.Cache__Service                                       import Cache__Service
from mgraph_ai_service_cache_client.schemas.cache.zip.Schema__Cache__Zip__Operation__Request    import Schema__Cache__Zip__Operation__Request
from mgraph_ai_service_cache_client.schemas.cache.zip.Schema__Cache__Zip__Operation__Response   import Schema__Cache__Zip__Operation__Response
from mgraph_ai_service_cache.utils.for_osbot_utils.Zip                                          import zip_bytes__content_hash


class Cache__Service__Zip__Operations(Type_Safe):                                        # Service layer for zip file operations
    cache_service : Cache__Service                                                       # Underlying cache service

    @cache_on_self
    def operation_handlers(self):
        return { Enum__Cache__Zip__Operation.LIST    : self.operation__list    ,
                 Enum__Cache__Zip__Operation.GET     : self.operation__get     ,
                 Enum__Cache__Zip__Operation.ADD     : self.operation__add     ,
                 Enum__Cache__Zip__Operation.REMOVE  : self.operation__remove  ,
                 Enum__Cache__Zip__Operation.REPLACE : self.operation__replace }

    @type_safe
    def perform_operation(self, request: Schema__Cache__Zip__Operation__Request
                           ) -> Schema__Cache__Zip__Operation__Response:                   # Route operation to appropriate handler
        try:
            handler = self.operation_handlers().get(request.operation)
            if not handler:
                return Schema__Cache__Zip__Operation__Response(success       = False                                    ,
                                                               operation     = request.operation                        ,
                                                               cache_id      = request.cache_id                         ,
                                                               error_details = f"Unknown operation: {request.operation}")

            response = handler(request)
            return response
        except Exception as e:
            return self.error_response(request, str(e))

    @type_safe
    def operation__list(self, request: Schema__Cache__Zip__Operation__Request
                         ) -> Schema__Cache__Zip__Operation__Response:                     # List all files in zip (read-only)
        zip_bytes = self.retrieve_zip_bytes(request.cache_id, request.namespace)
        if not zip_bytes:
            return self.error_response(request, "Zip file not found in cache")

        file_list = zip_bytes__file_list(zip_bytes)
        return Schema__Cache__Zip__Operation__Response(success    = True                                  ,
                                                       operation  = Enum__Cache__Zip__Operation.LIST      ,
                                                       cache_id   = request.cache_id                      ,
                                                       file_list  = file_list                             ,
                                                       message    = f"Found {len(file_list)} files in zip")

    @type_safe
    def operation__get(self, request: Schema__Cache__Zip__Operation__Request
                        ) -> Schema__Cache__Zip__Operation__Response:                      # Get single file from zip (read-only)
        if not request.file_path:
            return self.error_response(request, "file_path required for get operation")

        zip_bytes = self.retrieve_zip_bytes(request.cache_id, request.namespace)
        if not zip_bytes:
            return self.error_response(request, "Zip file not found in cache")

        file_content = zip_bytes__file(zip_bytes, str(request.file_path))
        if file_content is None:
            return self.error_response(request, f"File '{request.file_path}' not found in zip")

        return Schema__Cache__Zip__Operation__Response(success      = True                                ,
                                                       operation    = Enum__Cache__Zip__Operation.GET     ,
                                                       cache_id     = request.cache_id                    ,
                                                       file_content = file_content                        ,
                                                       file_size    = len(file_content)                   ,
                                                       message      = f"Retrieved '{request.file_path}'"  )

    @type_safe
    def operation__add(self, request: Schema__Cache__Zip__Operation__Request
                        ) -> Schema__Cache__Zip__Operation__Response:                           # Add new file to zip (creates new cache entry)
        if not request.file_path: #or not request.file_content:
            return self.error_response(request, "file_path required for get operation")

        zip_bytes = self.retrieve_zip_bytes(request.cache_id, request.namespace)
        if not zip_bytes:
            return self.error_response(request, "Zip file not found in cache")

        new_zip = zip_bytes__add_file(zip_bytes, str(request.file_path), request.file_content)

        new_cache_id = self.create_modified_zip(original_id = request.cache_id                      ,           # Create new immutable cache entry
                                                namespace   = request.namespace                     ,
                                                zip_bytes   = new_zip                               ,
                                                operation   = "add"                                 ,
                                                details     = {"added_file": str(request.file_path)})

        if not new_cache_id:
            return self.error_response(request, "Failed to create new cache entry")

        return Schema__Cache__Zip__Operation__Response(success           = True                                               ,
                                                       operation         = Enum__Cache__Zip__Operation.ADD                    ,
                                                       cache_id          = new_cache_id                                       ,  # NEW cache ID
                                                       original_cache_id = request.cache_id                                   ,  # Original for provenance
                                                       files_affected    = [request.file_path]                                ,
                                                       message           = f"Added '{request.file_path}' to zip (new cache_id: {new_cache_id})")

    @type_safe
    def operation__remove(self, request: Schema__Cache__Zip__Operation__Request
                         ) -> Schema__Cache__Zip__Operation__Response:                   # Remove file from zip (creates new cache entry)
        if not request.file_path:
            return self.error_response(request, "file_path required for remove operation")

        zip_bytes = self.retrieve_zip_bytes(request.cache_id, request.namespace)
        if not zip_bytes:
            return self.error_response(request, "Zip file not found in cache")

        new_zip = zip_bytes__remove_file(zip_bytes, str(request.file_path))

        # Create new immutable cache entry
        new_cache_id = self.create_modified_zip(original_id = request.cache_id                                 ,
                                                namespace   = request.namespace                                ,
                                                zip_bytes   = new_zip                                          ,
                                                operation   = "remove"                                         ,
                                                details     = {"removed_file": str(request.file_path)})

        if not new_cache_id:
            return self.error_response(request, "Failed to create new cache entry")

        return Schema__Cache__Zip__Operation__Response(
            success           = True                                                 ,
            operation         = Enum__Cache__Zip__Operation.REMOVE                   ,
            cache_id          = new_cache_id                                         ,  # NEW cache ID
            original_cache_id = request.cache_id                                     ,  # Original for provenance
            files_affected    = [request.file_path]                                  ,
            message           = f"Removed '{request.file_path}' from zip (new cache_id: {new_cache_id})")

    @type_safe
    def operation__replace(self, request: Schema__Cache__Zip__Operation__Request
                          ) -> Schema__Cache__Zip__Operation__Response:                  # Replace existing file in zip (creates new cache entry)
        if not request.file_path:
            return self.error_response(request, "file_path required for replace operation")

        zip_bytes = self.retrieve_zip_bytes(request.cache_id, request.namespace)
        if not zip_bytes:
            return self.error_response(request, "Zip file not found in cache")

        new_zip = zip_bytes__replace_file(zip_bytes, str(request.file_path), request.file_content)

        # Create new immutable cache entry
        new_cache_id = self.create_modified_zip(original_id = request.cache_id                                  ,
                                                namespace   = request.namespace                                 ,
                                                zip_bytes   = new_zip                                           ,
                                                operation   = "replace"                                         ,
                                                details     = {"replaced_file": str(request.file_path)}
)

        if not new_cache_id:
            return self.error_response(request, "Failed to create new cache entry")

        return Schema__Cache__Zip__Operation__Response(
            success           = True                                                   ,
            operation         = Enum__Cache__Zip__Operation.REPLACE                    ,
            cache_id          = new_cache_id                                           ,  # NEW cache ID
            original_cache_id = request.cache_id                                       ,  # Original for provenance
            files_affected    = [request.file_path]                                    ,
            message           = f"Replaced '{request.file_path}' in zip (new cache_id: {new_cache_id})")

    def create_modified_zip(self, original_id : Cache_Id     ,
                                  namespace   : str          ,
                                  zip_bytes   : bytes        ,
                                  operation   : str          ,
                                  details     : dict = None
                             ) -> Optional[Cache_Id   ]:                                                            # Create new immutable cache entry for modified zip

        hash_length = self.cache_service.hash_config.length
        cache_hash  = zip_bytes__content_hash(zip_bytes   = zip_bytes,
                                              hash_length = hash_length)                                            # use the bytes content and files path to calculate the zip's hash (which is different mode than when we just store bytes)

        original_entry__refs      = self.cache_service.retrieve_by_id__refs    (original_id, namespace)             # Get original entry to preserve strategy
        #original_entry__metadata  = self.cache_service.retrieve_by_id__metadata(original_id, namespace)           # Get original entry to preserve strategy
        original_strategy = original_entry__refs.strategy

        # todo: wire up the update of the metadata  (note: need to figure out the best place to put it
        # metadata__original = original_entry__metadata.data
        # metadata = { **metadata__original                  ,                            # Build provenance metadata
        #              "parent_id"      : str(original_id)   ,
        #              "operation"      : operation          ,
        #              "transformation" : f"zip_{operation}" ,
        #              "derived_from"   : str(original_id)   }
        # if details:
        #     metadata["operation_details"] = details

        result = self.cache_service.store_with_strategy(storage_data = zip_bytes              ,  # Store as new immutable entry with same strategy as original
                                                        cache_hash   = cache_hash             ,
                                                        cache_id     = Cache_Id(Random_Guid()),   # Always new ID for immutability
                                                        namespace    = namespace              ,
                                                        strategy     = original_strategy)   # Preserve original's strategy
        return result.cache_id if result else None

    def retrieve_zip_bytes(self, cache_id  : Random_Guid    ,
                                 namespace : str
                          ) -> Optional[bytes]:                                          # Get zip from cache
        result = self.cache_service.retrieve_by_id(cache_id, namespace)
        if result and result.get('data_type') == 'binary':
            return result.get('data')
        return None

    def error_response(self, request       : Schema__Cache__Zip__Operation__Request,
                             error_details : str
                       ) -> Schema__Cache__Zip__Operation__Response:                    # Create error response
        return Schema__Cache__Zip__Operation__Response(
            success       = False                 ,
            operation     = request.operation     ,
            cache_id      = request.cache_id      ,
            error_details = error_details         )


from fastapi                                                                                      import HTTPException
from osbot_fast_api.api.decorators.route_path                                                     import route_path
from osbot_fast_api.api.routes.Fast_API__Routes                                                   import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix                        import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag                           import Safe_Str__Fast_API__Route__Tag
from osbot_utils.decorators.methods.cache_on_self                                                 import cache_on_self
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path                 import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                                import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id                   import Safe_Str__Id
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__Retrieve__Request     import Schema__Cache__Data__Retrieve__Request
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Data_Type                    import Enum__Cache__Data_Type
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API                                import FAST_API__PARAM__NAMESPACE
from mgraph_ai_service_cache.service.cache.Cache__Service                                         import Cache__Service
from mgraph_ai_service_cache.service.cache.data.Cache__Service__Data__Delete                      import Cache__Service__Data__Delete

TAG__ROUTES_DELETE__DATA       = Safe_Str__Fast_API__Route__Tag('data')
PREFIX__ROUTES_DELETE__DATA    = Safe_Str__Fast_API__Route__Prefix('/{namespace}/cache/{cache_id}')
BASE_PATH__ROUTES_DELETE__DATA = f'{PREFIX__ROUTES_DELETE__DATA}/{TAG__ROUTES_DELETE__DATA}/'

ROUTES_PATHS__DELETE__DATA = [ BASE_PATH__ROUTES_DELETE__DATA + 'delete/{data_type}/{data_file_id}'                  ,
                               BASE_PATH__ROUTES_DELETE__DATA + 'delete/{data_type}/{data_key:path}/{data_file_id}' ,
                               BASE_PATH__ROUTES_DELETE__DATA + 'delete/all'                                        ,
                               BASE_PATH__ROUTES_DELETE__DATA + 'delete/all/{data_key:path}'                        ]


class Routes__Data__Delete(Fast_API__Routes):                                                   # FastAPI routes for deleting data files from cache entries
    tag           : Safe_Str__Fast_API__Route__Tag    = TAG__ROUTES_DELETE__DATA
    prefix        : Safe_Str__Fast_API__Route__Prefix = PREFIX__ROUTES_DELETE__DATA
    cache_service : Cache__Service                                                              # Dependency injection for cache service

    @cache_on_self
    def delete_service(self) -> Cache__Service__Data__Delete:                                   # Service layer for delete operations
        return Cache__Service__Data__Delete(cache_service=self.cache_service)

    @route_path("/data/delete/{data_type}/{data_file_id}")
    def delete__data__file__with__id(self, cache_id     : Cache_Id               = None                          ,
                                           namespace    : Safe_Str__Id           = FAST_API__PARAM__NAMESPACE    ,
                                           data_type    : Enum__Cache__Data_Type = None                          ,
                                           data_file_id : Safe_Str__Id           = None
                                     ) -> dict:
        return self.delete__data__file__with__id_and_key(cache_id     = cache_id    ,
                                                         namespace    = namespace   ,
                                                         data_key     = ''          ,
                                                         data_type    = data_type   ,
                                                         data_file_id = data_file_id)

    @route_path("/data/delete/{data_type}/{data_key:path}/{data_file_id}")
    def delete__data__file__with__id_and_key(self, cache_id     : Cache_Id               = None                          ,
                                                   namespace    : Safe_Str__Id           = FAST_API__PARAM__NAMESPACE    ,
                                                   data_type    : Enum__Cache__Data_Type = None                          ,
                                                   data_key     : Safe_Str__File__Path   = None                          ,
                                                   data_file_id : Safe_Str__Id           = None
                                             ) -> dict:                                                                   # Delete specific data file

        if not data_file_id:
            error_detail = { "error_type" : "INVALID_INPUT",
                            "message"    : "data_file_id is required to delete a specific data file"}
            raise HTTPException(status_code=400, detail=error_detail)

        if not data_type:
            error_detail = { "error_type" : "INVALID_INPUT",
                            "message"    : "data_type is required to delete a specific data file"}
            raise HTTPException(status_code=400, detail=error_detail)

        request = Schema__Cache__Data__Retrieve__Request(cache_id     = cache_id    ,
                                                         data_type    = data_type   ,
                                                         data_key     = data_key    ,
                                                         data_file_id = data_file_id,
                                                         namespace    = namespace   )

        deleted = self.delete_service().delete_data_file(request)

        if not deleted:
            error_detail = { "error_type"   : "NOT_FOUND"                                       ,           # todo: refactor to use Type_Safe classes
                             "message"      : f"Data file {data_file_id} not found"             ,
                             "cache_id"     : str(cache_id)                                     ,
                             "data_file_id" : str(data_file_id)                                 ,
                             "data_type"    : data_type.value                                   }
            raise HTTPException(status_code=404, detail=error_detail)

        return { "status"        : "success"                                                    ,           # todo: refactor to use Type_Safe classes
                "message"       : "Data file deleted successfully"                             ,
                "cache_id"      : str(cache_id)                                                ,
                "data_file_id"  : str(data_file_id)                                            ,
                "data_type"     : data_type.value                                              ,
                "data_key"      : str(data_key) if data_key else None                          ,
                "namespace"     : str(namespace)                                               }

    @route_path("/data/delete/all")
    def delete__all__data__files(self, cache_id  : Cache_Id    = None                          ,
                                       namespace : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                ) -> dict:
        return self.delete__all__data__files__with__key(cache_id  = cache_id ,
                                                        namespace = namespace,
                                                        data_key  = ''       )

    @route_path("/data/delete/all/{data_key:path}")
    def delete__all__data__files__with__key(self, cache_id  : Cache_Id             = None                          ,
                                                  namespace : Safe_Str__Id         = FAST_API__PARAM__NAMESPACE    ,
                                                  data_key  : Safe_Str__File__Path = None
                                           ) -> dict:                                                               # Delete all data files
        result = self.delete_service().delete_all_data_files(cache_id  = cache_id ,
                                                             namespace = namespace,
                                                             data_key  = data_key )

        deleted_count = result.deleted_count

        if deleted_count == 0:
            return { "status"        : "success"                                                ,               # todo: refactor to use Type_Safe classes
                    "message"       : "No data files to delete"                                ,
                    "cache_id"      : str(cache_id)                                            ,
                    "deleted_count" : 0                                                        ,
                    "data_key"      : str(data_key) if data_key else None                      ,
                    "namespace"     : str(namespace)                                           }

        return { "status"        : "success"                                                   ,
                "message"       : f"Deleted {deleted_count} data files"                       ,                 # todo: refactor to use Type_Safe classes
                "cache_id"      : str(cache_id)                                               ,
                "deleted_count" : deleted_count                                               ,
                "deleted_files" : result.deleted_files                                        ,
                "data_key"      : str(data_key) if data_key else None                         ,
                "namespace"     : str(namespace)                                              }

    def setup_routes(self):                                                                     # Configure all data deletion routes

        self.add_route_delete(self.delete__all__data__files               )                 # IMPORTANT: from a routing point of view, this needs to be added before the delete__data__file__with__id
        self.add_route_delete(self.delete__all__data__files__with__key    )                 #            so that the /all/ is picked up
        self.add_route_delete(self.delete__data__file__with__id           )
        self.add_route_delete(self.delete__data__file__with__id_and_key   )


from memory_fs.Memory_FS                                                                import Memory_FS
from memory_fs.helpers.Memory_FS__Temporal                                              import Memory_FS__Temporal
from memory_fs.helpers.Memory_FS__Latest_Temporal                                       import Memory_FS__Latest_Temporal
from memory_fs.path_handlers.Path__Handler__Hash_Sharded                                import Path__Handler__Hash_Sharded
from memory_fs.path_handlers.Path__Handler__Key_Based                                   import Path__Handler__Key_Based
from memory_fs.storage_fs.Storage_FS                                                    import Storage_FS
from osbot_utils.type_safe.Type_Safe                                                    import Type_Safe
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path       import Safe_Str__File__Path
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Store__Strategy    import Enum__Cache__Store__Strategy
from osbot_utils.utils.Http                                                             import url_join_safe

# Constants for all prefix paths
CACHE__HANDLER__PREFIX_PATH__FS__REFS_ID                            = 'refs/by-id'
CACHE__HANDLER__PREFIX_PATH__FS__REFS_HASH                          = 'refs/by-hash'
CACHE__HANDLER__PREFIX_PATH__FS__DATA_DIRECT                        = 'data/direct'
CACHE__HANDLER__PREFIX_PATH__FS__DATA_KEY_BASED                     = 'data/key-based'
CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL                      = 'data/temporal'
CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL_LATEST               = 'data/temporal-latest'
CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL_VERSIONED            = 'data/temporal-versioned'
CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL_VERSIONED_VERSIONS   = 'data/temporal-versioned/versions'

# Default shard depth for hash sharding
CACHE__HANDLER__SHARD_DEPTH_DEFAULT = 2

class Cache__Handler(Type_Safe):                                                                        # Single handler supporting all storage strategies
    storage_backend         : Storage_FS                  = None                                        # Storage backend instance (from Cache__Config)
    namespace               : str                         = ""                                          # Namespace prefix for isolation
    cache_ttl_hours         : int                         = 24

    # All Memory_FS instances for different strategies
    fs__data_direct             : Memory_FS                   = None                                    # Direct to hash location
    fs__data_temporal           : Memory_FS__Temporal         = None                                    # Temporal organization
    fs__data_temporal_latest    : Memory_FS__Latest_Temporal  = None                                    # Temporal + latest
    fs__data_temporal_versioned : Memory_FS                   = None                                    # Temporal + latest + versions
    fs__data_key_based          : Memory_FS                   = None                                    # For Semantic Files

    # Reference stores (always needed)
    fs__refs_hash               : Memory_FS                   = None                                        # Hash->ID mappings
    fs__refs_id                 : Memory_FS                   = None                                        # ID->Hash mappings

    def build_namespaced_path(self, base_path: str) -> Safe_Str__File__Path:                           # Helper to build namespace-prefixed paths
        if self.namespace:                                                                              # Build a path with namespace prefix if namespace is set
            return url_join_safe(self.namespace , Safe_Str__File__Path(base_path))
        return Safe_Str__File__Path(base_path)

    def setup(self) -> 'Cache__Handler':                                                                # Setup all storage strategies
        if not self.storage_backend:
            raise ValueError("Storage backend must be provided before setup")

        storage = self.storage_backend                                                                  # Use the storage backend directly - namespace is handled via path prefixes

        # todo: refactor the creation of these stores into separate methods (one method per handler)
        # Reference stores (always needed) - with namespace prefix
        with Memory_FS() as _:                                                                          # Hash->ID mappings
            kwargs__fs__refs_hash = dict(prefix_path = self.build_namespaced_path(CACHE__HANDLER__PREFIX_PATH__FS__REFS_HASH),
                                         shard_depth = CACHE__HANDLER__SHARD_DEPTH_DEFAULT)
            _.storage_fs = storage
            _.add_handler(Path__Handler__Hash_Sharded(**kwargs__fs__refs_hash))
            self.fs__refs_hash = _

        with Memory_FS() as _:                                                                          # ID->Hash mappings
            kwargs__fs__refs_id = dict(prefix_path = self.build_namespaced_path(CACHE__HANDLER__PREFIX_PATH__FS__REFS_ID),
                                       shard_depth = CACHE__HANDLER__SHARD_DEPTH_DEFAULT)
            _.storage_fs = storage
            _.add_handler(Path__Handler__Hash_Sharded(**kwargs__fs__refs_id))
            self.fs__refs_id = _

        with Memory_FS() as _:                                                                          # Direct strategy
            kwargs__fs__data_direct = dict(prefix_path = self.build_namespaced_path(CACHE__HANDLER__PREFIX_PATH__FS__DATA_DIRECT),
                                           shard_depth = CACHE__HANDLER__SHARD_DEPTH_DEFAULT)
            _.storage_fs = storage
            _.add_handler(Path__Handler__Hash_Sharded(**kwargs__fs__data_direct))
            self.fs__data_direct = _


        prefix_path = self.build_namespaced_path(CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL)       # Temporal only - with namespace prefix
        with Memory_FS__Temporal(storage_fs=storage, prefix_path=prefix_path) as _:
            self.fs__data_temporal = _

        # Temporal + Latest - with namespace prefix
        prefix_path = self.build_namespaced_path(CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL_LATEST)
        with Memory_FS__Latest_Temporal(storage_fs=storage, prefix_path=prefix_path) as _:
            self.fs__data_temporal_latest = _

        # Temporal + Latest + Versioned - with namespace prefixes
        with Memory_FS() as _:
            _.storage_fs = storage
            _.add_handler__temporal (prefix_path = self.build_namespaced_path(CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL_VERSIONED))
            _.add_handler__latest   (prefix_path = self.build_namespaced_path(CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL_VERSIONED))
            _.add_handler__versioned(prefix_path = self.build_namespaced_path(CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL_VERSIONED_VERSIONS))
            self.fs__data_temporal_versioned = _

        # Semantic File - with namespace prefix
        with Memory_FS() as _:
            kwargs__fs__data_semantic = dict(prefix_path = self.build_namespaced_path(CACHE__HANDLER__PREFIX_PATH__FS__DATA_KEY_BASED))
            _.storage_fs = storage
            _.add_handler(Path__Handler__Key_Based(**kwargs__fs__data_semantic))
            self.fs__data_key_based = _

        return self

    def get_fs_for_strategy(self, strategy: Enum__Cache__Store__Strategy
                             ) -> Memory_FS:                                                                         # Return appropriate Memory_FS for strategy
        if strategy == "direct":
            return self.fs__data_direct
        elif strategy == "temporal":
            return self.fs__data_temporal
        elif strategy == "temporal_latest":
            return self.fs__data_temporal_latest
        elif strategy == "temporal_versioned":
            return self.fs__data_temporal_versioned
        elif strategy == "key_based":
            return self.fs__data_key_based
        else:
            raise ValueError(f"Unknown strategy: {strategy}")


import gzip
from osbot_utils.type_safe.Type_Safe                                                            import Type_Safe
from osbot_utils.type_safe.primitives.domains.common.safe_str.Safe_Str__Text                    import Safe_Str__Text
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path               import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id                 import Safe_Str__Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Json__Field_Path   import Safe_Str__Json__Field_Path
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                                  import type_safe
from mgraph_ai_service_cache_client.schemas.cache.consts__Cache_Service                         import DEFAULT_CACHE__STORE__STRATEGY, DEFAULT_CACHE__NAMESPACE
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Store__Strategy            import Enum__Cache__Store__Strategy
from mgraph_ai_service_cache_client.schemas.errors.Schema__Cache__Error__Invalid_Input          import Schema__Cache__Error__Invalid_Input
from mgraph_ai_service_cache.service.cache.Cache__Service                                       import Cache__Service
from mgraph_ai_service_cache_client.schemas.cache.Schema__Cache__Store__Response                import Schema__Cache__Store__Response


class Cache__Service__Store(Type_Safe):                                               # Service layer for cache store operations
    cache_service : Cache__Service                                                    # Underlying cache service

    @type_safe
    def store_string(self, data      : str                                                          ,       # todo: we should we using a Type_Safe class for these params
                           strategy  : Enum__Cache__Store__Strategy = DEFAULT_CACHE__STORE__STRATEGY,
                           namespace : Safe_Str__Id                 = DEFAULT_CACHE__NAMESPACE      ,
                           cache_key : Safe_Str__File__Path         = None                          ,
                           file_id   : Safe_Str__Id                 = None
                     ) -> Schema__Cache__Store__Response:                               # Store string data

        if not data:                                                                    # Validate input
            return None                                                                 # Let route handler decide error response

        cache_hash = self.cache_service.hash_from_string(data)

        return self.cache_service.store_with_strategy(storage_data = data       ,       # todo: convert this into a Type_Safe object
                                                      cache_hash   = cache_hash ,
                                                      cache_key    = cache_key  ,
                                                      file_id      = file_id    ,
                                                      strategy     = strategy   ,
                                                      namespace    = namespace  )

    @type_safe
    def store_json(self, data             : dict                         = None                          ,
                         strategy         : Enum__Cache__Store__Strategy = DEFAULT_CACHE__STORE__STRATEGY,
                         namespace        : Safe_Str__Id                 = DEFAULT_CACHE__NAMESPACE      ,
                         cache_key        : Safe_Str__File__Path         = None                          ,
                         file_id          : Safe_Str__Id                 = None                          ,
                         json_field_path  : Safe_Str__Json__Field_Path   = None                          ,
                    ) -> Schema__Cache__Store__Response:                                                # Store JSON data

        if json_field_path:                                                                             # Field-based hashing
            cache_hash = self.cache_service.hash_from_json_field(data=data, json_field=json_field_path)
        else:
            cache_hash = self.cache_service.hash_from_json(data)

        return self.cache_service.store_with_strategy(storage_data    = data           ,        # todo: review refactoring opportunity with store_string since a lot of the code in this method is very similar
                                                      cache_hash      = cache_hash     ,
                                                      cache_key       = cache_key      ,
                                                      file_id         = file_id        ,
                                                      json_field_path = json_field_path,
                                                      strategy        = strategy       ,
                                                      namespace       = namespace      )

    @type_safe
    def store_binary(self, data             : bytes,
                           strategy          : Enum__Cache__Store__Strategy = DEFAULT_CACHE__STORE__STRATEGY,
                           namespace         : Safe_Str__Id                 = DEFAULT_CACHE__NAMESPACE      ,
                           cache_key         : Safe_Str__File__Path         = None                          ,
                           file_id           : Safe_Str__Id                 = None                          ,
                           content_encoding  : str                          = None
                      ) -> Schema__Cache__Store__Response:                               # Store binary data

        if not data:                                                                    # Validate input
            return None                                                                 # Let route handler decide error response

        # todo: review the move of this gzip capability into a separate store_binary__gzip method (to keep the paths clean
        if content_encoding == 'gzip':                                                  # Handle compression
            decompressed = gzip.decompress(data)                                        # get decompressed data so that we can use it to calculate the hash
            cache_hash = self.cache_service.hash_from_bytes(decompressed)               # todo: question: if we are storing the compressed bytes, shouldn't the hash be calculated from those compressed bytes?
            storage_data = data                                                         # Store compressed
        else:
            cache_hash = self.cache_service.hash_from_bytes(data)
            storage_data = data


        return self.cache_service.store_with_strategy(storage_data     = storage_data    ,  # todo: when looking at the refactoring opportunities, since most of these values are being passed through
                                                      cache_hash       = cache_hash      ,  #       these store_* functions, it would be better to have a domain class that would hold those values
                                                      cache_key        = cache_key       ,  #          a) don't fit in that domain class
                                                      file_id          = file_id         ,  #          b) have been modified in this function and need to be passed to the next function
                                                      strategy         = strategy        ,
                                                      namespace        = namespace       ,
                                                      content_encoding = content_encoding )


    @type_safe
    def get_invalid_input_error(self, field_name    : Safe_Str__Id,
                                      field_value   : Safe_Str__Text = None,
                                      expected_type : Safe_Str__Id   = None,
                                      message       : Safe_Str__Text = None
                                 ) -> Schema__Cache__Error__Invalid_Input:                                                       # Build invalid input error
        return Schema__Cache__Error__Invalid_Input( error_type    = "INVALID_INPUT"                                       ,     # todo: refactor to error type to Enum
                                                    message       = message or f"Invalid value for field {field_name}"    ,
                                                    field_name    = field_name                                            ,
                                                    field_value   = str(field_value) if field_value else None             ,
                                                    expected_type = expected_type                                         )


from typing                                                                                     import Union, Dict
from osbot_utils.type_safe.Type_Safe                                                            import Type_Safe
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash        import Safe_Str__Cache_Hash
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path               import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                              import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_int.Timestamp_Now                import Timestamp_Now
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id                 import Safe_Str__Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Json__Field_Path   import Safe_Str__Json__Field_Path
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Store__Strategy            import Enum__Cache__Store__Strategy
from mgraph_ai_service_cache_client.schemas.cache.file.Schema__Cache__File__Paths               import Schema__Cache__File__Paths
from mgraph_ai_service_cache_client.schemas.cache.store.Schema__Cache__Store__Metadata          import Schema__Cache__Store__Metadata
from mgraph_ai_service_cache_client.schemas.cache.store.Schema__Cache__Store__Paths             import Schema__Cache__Store__Paths

# todo: review the use of Cache__Handler since it is this dependency that makes this entire schema to need to be placed in the mgraph_ai_service_cache project (instead of the mgraph_ai_service_cache__client project)
from mgraph_ai_service_cache.service.cache.Cache__Handler                                   import Cache__Handler


class Schema__Store__Context(Type_Safe):                                                      # Context object to pass data between store operations
    storage_data      : Union[str, Dict, bytes]                                               # Data to be stored (string, dict, or bytes)
    cache_hash        : Safe_Str__Cache_Hash         = None                                   # Hash of the data or cache key
    cache_id          : Cache_Id                     = None                                   # Unique ID for this cache entry
    cache_key         : Safe_Str__File__Path         = None                                   # Optional semantic cache key
    file_id           : Safe_Str__Id                 = None                                   # Optional file ID (defaults to cache_id)
    json_field_path   : Safe_Str__Json__Field_Path   = None                                   # Field Path used to calculate the hash of a Json object
    namespace         : Safe_Str__Id                 = None                                   # Namespace for isolation
    strategy          : Enum__Cache__Store__Strategy = None                                   # Storage strategy to use
    content_encoding  : Safe_Str__Id                 = None                                   # Optional encoding (e.g., 'gzip')
    handler           : Cache__Handler               = None                                   # Cache handler for the namespace

    # Computed during storage process (now using Type_Safe classes)
    file_type          : str                              = None                               # Determined type: 'json' or 'binary'
    file_size          : int                              = None                               # Size of stored data in bytes
    all_paths          : Schema__Cache__Store__Paths      = None                               # Paths organized by type
    file_paths         : Schema__Cache__File__Paths                                            # Paths to actual content files and data folders
    timestamp          : Timestamp_Now                    = None                               # When the entry was stored
    metadata           : Schema__Cache__Store__Metadata   = None


from osbot_utils.type_safe.primitives.domains.common.safe_str.Safe_Str__Version import Safe_Str__Version
from osbot_utils.type_safe.Type_Safe                                 import Type_Safe
from osbot_utils.utils.Version                                       import Version as Version__OSBot_Utils
from osbot_aws.utils.Version                                         import Version as Version__OSBot_AWS
from osbot_fast_api.utils.Version                                    import version__osbot_fast_api
from osbot_fast_api_serverless.utils.Version                         import version__osbot_fast_api_serverless
from mgraph_ai_service_cache.utils.Version                            import version__mgraph_ai_service_cache


class Schema__Server__Versions(Type_Safe):
    mgraph_ai_service_cache   : Safe_Str__Version       = version__mgraph_ai_service_cache
    osbot_utils              : Safe_Str__Version       = Safe_Str__Version(Version__OSBot_Utils().value()    )
    osbot_aws                : Safe_Str__Version       = Safe_Str__Version(Version__OSBot_AWS  ().value()    )
    osbot_fast_api           : Safe_Str__Version       = Safe_Str__Version(version__osbot_fast_api           )
    osbot_fast_api_serverless: Safe_Str__Version        =Safe_Str__Version(version__osbot_fast_api_serverless)



from typing                                                                                 import Optional
from osbot_utils.decorators.methods.cache_on_self                                           import cache_on_self
from osbot_utils.type_safe.Type_Safe                                                        import Type_Safe
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path           import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.core.Safe_UInt                                        import Safe_UInt
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                          import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id             import Safe_Str__Id
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                              import type_safe
from osbot_utils.utils.Http                                                                 import url_join_safe
from osbot_utils.utils.Json                                                                 import json_to_bytes
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__Files__List            import Schema__Cache__Data__Files__List
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__Retrieve__Request      import Schema__Cache__Data__Retrieve__Request
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__Retrieve__Response     import Schema__Cache__Data__Retrieve__Response
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Data_Type                     import Enum__Cache__Data_Type
from mgraph_ai_service_cache.service.cache.Cache__Service                                   import Cache__Service
from mgraph_ai_service_cache.service.cache.retrieve.Cache__Service__Retrieve                import Cache__Service__Retrieve


class Cache__Service__Data__Retrieve(Type_Safe):                                                # Service layer for retrieving data files from cache entries
    cache_service : Cache__Service

    @cache_on_self
    def retrieve_service(self):
        return Cache__Service__Retrieve(cache_service = self.cache_service)

    @type_safe
    def build_data_file_path(self, data_folder  : Safe_Str__File__Path,                         # Centralized path construction matching Store pattern
                                   data_key     : Safe_Str__File__Path = None,
                                   data_file_id : Safe_Str__Id         = None,
                                   data_type    : Enum__Cache__Data_Type = None
                            ) -> Safe_Str__File__Path:

        if not data_file_id:                                                                    # data_file_id is now required
            raise ValueError("data_file_id is required for data file operations")

        path = data_folder                                                                      # Start with base data folder

        if data_key:                                                                            # Add optional data_key path segment
            path = url_join_safe(path, data_key)

        path = url_join_safe(path, data_file_id)                                                # Add data_file_id

        if data_type:                                                                           # Add extension based on data type
            extension = self.get_extension_for_type(data_type)
            path = Safe_Str__File__Path(f"{path}.{extension}")

        return path

    @type_safe
    def retrieve_data(self, request: Schema__Cache__Data__Retrieve__Request
                       ) -> Schema__Cache__Data__Retrieve__Response:                             # Main retrieval method with explicit data type

        if not request.data_file_id:                                                           # Validate required fields
            return Schema__Cache__Data__Retrieve__Response(found=False,
                                                          data_type=request.data_type)

        if not request.data_type:
            raise ValueError("data_type is required for retrieval")

        file_refs = self.retrieve_service().retrieve_by_id__refs(cache_id  = request.cache_id  ,
                                                                 namespace = request.namespace)

        if not file_refs or not file_refs.file_paths.data_folders:
            return Schema__Cache__Data__Retrieve__Response(found=False,
                                                          data_type=request.data_type)

        handler = self.cache_service.get_or_create_handler(request.namespace)

        for data_folder in file_refs.file_paths.data_folders:                                   # Try each data folder
            full_path = self.build_data_file_path(data_folder       = data_folder      ,
                                                  data_key          = request.data_key ,
                                                  data_file_id      = request.data_file_id,
                                                  data_type         = request.data_type)

            data = self.read_file_by_type(storage_backend = handler.storage_backend,
                                          path            = full_path               ,
                                          data_type       = request.data_type      )

            if data is not None:                                                                # Found the file
                size = self.calculate_size(data, request.data_type)

                return Schema__Cache__Data__Retrieve__Response(data         = data                  ,
                                                               data_type    = request.data_type     ,
                                                               data_file_id = request.data_file_id  ,
                                                               data_key     = request.data_key      ,
                                                               full_path    = full_path             ,
                                                               size         = Safe_UInt(size)       ,
                                                               found        = True                  )

        return Schema__Cache__Data__Retrieve__Response(found     = False,
                                                       data_type = request.data_type)

    @type_safe
    def list_data_files(self, cache_id  : Cache_Id             ,                                # List all data files
                              namespace : Safe_Str__Id         ,
                              data_key  : Safe_Str__File__Path = None
                        ) -> Schema__Cache__Data__Files__List:

        file_refs = self.retrieve_service().retrieve_by_id__refs(cache_id = cache_id ,
                                                                 namespace = namespace)
        files_list = Schema__Cache__Data__Files__List(cache_id   = cache_id  ,
                                                      namespace  = namespace ,
                                                      data_key   = data_key  )
        if file_refs:
            data_folders = file_refs.file_paths.data_folders
            if data_folders:
                handler     = self.cache_service.get_or_create_handler(namespace)
                for data_folder in data_folders:
                    folder_path = data_folder
                    if data_key:                                                                     # Filter by data_key if provided
                        folder_path = url_join_safe(folder_path, data_key)
                    files_list.data_files = handler.storage_backend.folder__files__all(parent_folder=folder_path) # note: for now only return the list of files

        return files_list

    def read_file_by_type(self, storage_backend,                                                # Read file based on explicit type
                               path     : Safe_Str__File__Path,
                               data_type: Enum__Cache__Data_Type) -> Optional[any]:

        if not storage_backend.file__exists(path):
            return None

        if data_type == Enum__Cache__Data_Type.JSON:
            return storage_backend.file__json(path)
        elif data_type == Enum__Cache__Data_Type.STRING:
            return storage_backend.file__str(path)
        elif data_type == Enum__Cache__Data_Type.BINARY:
            return storage_backend.file__bytes(path)
        else:
            raise ValueError(f"Unknown data type: {data_type}")

    def get_extension_for_type(self, data_type: Enum__Cache__Data_Type) -> str:                 # Extension mapping (same as Store)
        extensions = { Enum__Cache__Data_Type.STRING : 'txt' ,
                      Enum__Cache__Data_Type.JSON   : 'json',
                      Enum__Cache__Data_Type.BINARY : 'bin' }
        return extensions.get(data_type, "data")

    def get_type_from_extension(self, extension: str) -> Enum__Cache__Data_Type:                # Reverse mapping for listing
        ext_map = { 'json' : Enum__Cache__Data_Type.JSON  ,
                   'txt'  : Enum__Cache__Data_Type.STRING,
                   'bin'  : Enum__Cache__Data_Type.BINARY }
        return ext_map.get(extension, Enum__Cache__Data_Type.BINARY)

    def calculate_size(self, data, data_type: Enum__Cache__Data_Type) -> int:                   # Calculate data size based on type
        if data_type == Enum__Cache__Data_Type.BINARY:
            return len(data) if data else 0
        elif data_type == Enum__Cache__Data_Type.STRING:
            return len(data.encode('utf-8')) if data else 0
        elif data_type == Enum__Cache__Data_Type.JSON:
            return len(json_to_bytes(data)) if data else 0        


from typing                                                                       import List, Optional
from osbot_aws.AWS_Config                                                         import aws_config
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path import Safe_Str__File__Path
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                    import type_safe
from osbot_utils.utils.Http                                                       import url_join_safe
from osbot_utils.utils.Json                                                       import bytes_to_json
from osbot_aws.aws.s3.S3                                                          import S3
from memory_fs.storage_fs.Storage_FS                                              import Storage_FS

# todo: see if we should move this to the boto3-fs project
class Storage_FS__S3(Storage_FS):
    s3_bucket    : str                                                                  # S3 bucket name for storage
    s3_prefix    : str = ""                                                             # Optional prefix for all keys
    s3           : S3  = None                                                           # S3 instance (will be created if not provided)
    
    def setup(self) -> 'Storage_FS__S3':                                                # Initialize S3 client if not provided
        if self.s3 is None:
            self.s3 = S3()

        if not self.s3.bucket_exists(self.s3_bucket):                                   # Ensure bucket exists
            region = aws_config.region_name()                                           # Get region from AWS config
            result = self.s3.bucket_create(bucket=self.s3_bucket, region=region)
            if result.get('status') != 'ok':
                raise Exception(f"Failed to create S3 bucket {self.s3_bucket}: {result}")
        
        return self
    
    def _get_s3_key(self, path: Safe_Str__File__Path) -> str:                                       # Convert file path to S3 key with optional prefix
        key = str(path)
        if self.s3_prefix:
            prefix = self.s3_prefix if self.s3_prefix.endswith('/') else f"{self.s3_prefix}/"       # Ensure prefix ends with / if it's not empty
            key = f"{prefix}{key}"
        return key
    
    def _get_path_from_key(self, s3_key: str) -> Safe_Str__File__Path:                              # Convert S3 key back to path
        if self.s3_prefix:
            prefix = self.s3_prefix if self.s3_prefix.endswith('/') else f"{self.s3_prefix}/"
            if s3_key.startswith(prefix):
                s3_key = s3_key[len(prefix):]
        return Safe_Str__File__Path(s3_key)
    
    @type_safe
    def file__bytes(self, path: Safe_Str__File__Path                                                # Read file content as bytes from S3
                    ) -> Optional[bytes]:
        s3_key = self._get_s3_key(path)
        if self.file__exists(path):
            return self.s3.file_bytes(bucket=self.s3_bucket, key=s3_key)
        return None
    
    @type_safe
    def file__delete(self, path: Safe_Str__File__Path                                  # Delete a file from S3
                     ) -> bool:
        s3_key = self._get_s3_key(path)
        if self.file__exists(path) is True:
            return self.s3.file_delete(bucket=self.s3_bucket, key=s3_key)
        return False
    
    @type_safe
    def file__exists(self, path: Safe_Str__File__Path                                  # Check if file exists in S3
                     ) -> bool:
        s3_key = self._get_s3_key(path)
        return self.s3.file_exists(bucket=self.s3_bucket, key=s3_key)

    # todo: review why we are having issues with the return type
    @type_safe
    def file__json(self, path: Safe_Str__File__Path                                                 # Read file content as bytes from S3
                    ):     # : -> Optional[bytes]:          #   todo: review this usafe since it was cause some exceptions on the fast_api service
        file_bytes = self.file__bytes(path)
        if file_bytes:
            return bytes_to_json(file_bytes)
        return None
    
    @type_safe
    def file__save(self, path: Safe_Str__File__Path,                                   # Save bytes to S3
                         data: bytes
                   ) -> bool:
        s3_key = self._get_s3_key(path)
        return self.s3.file_create_from_bytes(
            file_bytes=data,
            bucket=self.s3_bucket,
            key=s3_key
        )
    
    @type_safe
    def file__str(self, path: Safe_Str__File__Path                                     # Read file content as string from S3
                  ) -> Optional[str]:
        s3_key = self._get_s3_key(path)
        if self.file__exists(path):
            return self.s3.file_contents(bucket=self.s3_bucket, key=s3_key)
        return None
    
    def files__paths(self) -> List[Safe_Str__File__Path]:                              # List all file paths in S3 bucket
        # Use find_files to get all keys with the prefix
        prefix = self.s3_prefix if self.s3_prefix else ''
        s3_keys = self.s3.find_files(bucket=self.s3_bucket, prefix=prefix)
        
        # Convert S3 keys back to paths
        paths = []
        for s3_key in s3_keys:
            path = self._get_path_from_key(s3_key)
            paths.append(path)
        
        return sorted(paths)
    
    def clear(self) -> bool:                                                           # Clear all files in the storage (within prefix)
        # Get all files with the current prefix
        prefix = self.s3_prefix if self.s3_prefix else ''
        s3_keys = self.s3.find_files(bucket=self.s3_bucket, prefix=prefix)
        
        if s3_keys:
            # Use files_delete for batch deletion
            return self.s3.files_delete(bucket=self.s3_bucket, keys=s3_keys)
        
        return True
    
    # Additional S3-specific methods
    
    def file__metadata(self, path: Safe_Str__File__Path) -> Optional[dict]:            # Get S3 file metadata
        s3_key = self._get_s3_key(path)
        if self.file__exists(path):
            return self.s3.file_metadata(bucket=self.s3_bucket, key=s3_key)
        return None
    
    def file__metadata_update(self, path: Safe_Str__File__Path,                        # Update S3 file metadata
                              metadata: dict
                              ) -> bool:
        s3_key = self._get_s3_key(path)
        if self.file__exists(path):
            result = self.s3.file_metadata_update(
                bucket=self.s3_bucket,
                key=s3_key,
                metadata=metadata
            )
            return result is not None
        return False
    
    def file__copy(self, source_path: Safe_Str__File__Path,                            # Copy file within S3
                         dest_path: Safe_Str__File__Path
                   ) -> bool:
        source_key = self._get_s3_key(source_path)
        dest_key = self._get_s3_key(dest_path)
        
        if self.file__exists(source_path):
            result = self.s3.file_copy(
                bucket_source=self.s3_bucket,
                key_source=source_key,
                bucket_destination=self.s3_bucket,
                key_destination=dest_key
            )
            return result is not None
        return False
    
    def file__move(self, source_path: Safe_Str__File__Path,                            # Move file within S3
                         dest_path: Safe_Str__File__Path
                   ) -> bool:
        source_key = self._get_s3_key(source_path)
        dest_key = self._get_s3_key(dest_path)
        
        if self.file__exists(source_path):
            return self.s3.file_move(src_bucket=self.s3_bucket,
                                     src_key=source_key,
                                     dest_bucket=self.s3_bucket,
                                     dest_key=dest_key )
        return False
    
    def file__size(self, path: Safe_Str__File__Path) -> Optional[int]:                 # Get file size in bytes
        s3_key = self._get_s3_key(path)
        if self.file__exists(path):
            details = self.s3.file_details(bucket=self.s3_bucket, key=s3_key)
            if details:
                return details.get('ContentLength')
        return None
    
    def file__last_modified(self, path: Safe_Str__File__Path) -> Optional[str]:        # Get last modified time
        s3_key = self._get_s3_key(path)
        if self.file__exists(path):
            details = self.s3.file_details(bucket=self.s3_bucket, key=s3_key)
            if details:
                last_modified = details.get('LastModified')
                if last_modified:
                    return last_modified.isoformat()
        return None

    def folder__folders(self, parent_folder='', return_full_path=False):
        kwargs = dict(s3_bucket        = self.s3_bucket  ,
                      parent_folder    = parent_folder   ,
                      return_full_path = return_full_path)
        return self.s3.folder_list(**kwargs)

    def folder__files__all(self, parent_folder: str):
        kwargs = dict(bucket    = self.s3_bucket  ,
                      prefix    = parent_folder   )
        return self.s3.find_files(**kwargs)

    def folder__files(self, folder_path: str,                                          # List files in a specific folder
                            return_full_path: bool = False
                      ) -> List[Safe_Str__File__Path]:
        # Combine prefix with folder path
        if self.s3_prefix:
            full_prefix = url_join_safe(str(self.s3_prefix), str(folder_path))
        else:
            full_prefix = folder_path
        
        # Use S3's folder_files method
        files = self.s3.folder_files(s3_bucket        =  self.s3_bucket,
                                     parent_folder    = full_prefix    ,
                                     return_full_path = True           ) # Always get full path from S3
        
        # Convert to Safe_Str__File__Path objects
        paths = []
        for file_key in files:
            if return_full_path:
                path = Safe_Str__File__Path(file_key)
            else:
                path = self._get_path_from_key(file_key)
            paths.append(path)
        
        return sorted(paths)
    
    def pre_signed_url(self, path: Safe_Str__File__Path,                               # Generate pre-signed URL for file access
                             operation: str = 'get_object',
                             expiration: int = 3600
                       ) -> Optional[str]:
        s3_key = self._get_s3_key(path)
        return self.s3.create_pre_signed_url(
            bucket_name=self.s3_bucket,
            object_name=s3_key,
            operation=operation,
            expiration=expiration
        )
    
    def bucket_versioning_enabled(self) -> bool:                                       # Check if versioning is enabled
        return self.s3.bucket_versioning__enabled(self.s3_bucket)
    
    def file__versions(self, path: Safe_Str__File__Path) -> Optional[list]:            # Get file versions (if versioning enabled)
        s3_key = self._get_s3_key(path)
        return self.s3.file_versions(bucket=self.s3_bucket, key=s3_key)


from osbot_fast_api.api.decorators.route_path import route_path
from starlette.responses import RedirectResponse
from starlette.staticfiles import StaticFiles

import mgraph_ai_service_cache__web_console
from mgraph_ai_service_cache.fast_api.routes.Routes__Namespaces             import Routes__Namespaces
from mgraph_ai_service_cache.fast_api.routes.data.Routes__Data__Delete      import Routes__Data__Delete
from mgraph_ai_service_cache.fast_api.routes.data.Routes__Data__Retrieve    import Routes__Data__Retrieve
from mgraph_ai_service_cache.fast_api.routes.data.Routes__Data__Store       import Routes__Data__Store
from mgraph_ai_service_cache.fast_api.routes.file.Routes__File__Update      import Routes__File__Update
from mgraph_ai_service_cache.fast_api.routes.test_data.Routes__Test_Data import Routes__Test_Data
from mgraph_ai_service_cache.fast_api.routes.zip.Routes__Zip                import Routes__Zip
from mgraph_ai_service_cache.service.cache.Cache__Service                   import Cache__Service
from osbot_fast_api.api.routes.Routes__Set_Cookie                           import Routes__Set_Cookie
from osbot_fast_api_serverless.fast_api.Serverless__Fast_API                import Serverless__Fast_API
from mgraph_ai_service_cache.config import FAST_API__TITLE, CACHE_SERVICE__WEB_CONSOLE__MAJOR__VERSION, CACHE_SERVICE__WEB_CONSOLE__LATEST__VERSION, CACHE_SERVICE__WEB_CONSOLE__ROUTE__START_PAGE, \
    CACHE_SERVICE__WEB_CONSOLE__PATH
from mgraph_ai_service_cache.fast_api.routes.file.Routes__File__Delete      import Routes__File__Delete
from mgraph_ai_service_cache.fast_api.routes.file.Routes__File__Exists      import Routes__File__Exists
from mgraph_ai_service_cache.fast_api.routes.file.Routes__File__Retrieve    import Routes__File__Retrieve
from mgraph_ai_service_cache.fast_api.routes.file.Routes__File__Store       import Routes__File__Store
from mgraph_ai_service_cache.fast_api.routes.Routes__Info                   import Routes__Info
from mgraph_ai_service_cache.fast_api.routes.Routes__Namespace              import Routes__Namespace
from mgraph_ai_service_cache.fast_api.routes.Routes__Server                 import Routes__Server
from mgraph_ai_service_cache.fast_api.routes.admin.Routes__Admin__Storage   import Routes__Admin__Storage
from mgraph_ai_service_cache.utils.Version                                  import version__mgraph_ai_service_cache

class Cache_Service__Fast_API(Serverless__Fast_API):
    cache_service: Cache__Service

    def setup(self):
        with self.config as _:
            _.title  = FAST_API__TITLE
            _.version = version__mgraph_ai_service_cache
        return super().setup()

    def setup_routes(self):
        self.add_routes(Routes__File__Store   )
        self.add_routes(Routes__File__Retrieve)
        self.add_routes(Routes__File__Exists  )
        self.add_routes(Routes__File__Update  )
        self.add_routes(Routes__File__Delete  )
        self.add_routes(Routes__Data__Store   )
        self.add_routes(Routes__Data__Retrieve)
        self.add_routes(Routes__Data__Delete  )
        self.add_routes(Routes__Zip           )
        self.add_routes(Routes__Namespace     )
        self.add_routes(Routes__Namespaces    )
        self.add_routes(Routes__Admin__Storage)
        self.add_routes(Routes__Server        )
        self.add_routes(Routes__Info          )
        self.add_routes(Routes__Set_Cookie    )
        self.add_routes(Routes__Test_Data     )
        self.setup_web_console()

    def add_routes(self, class_routes):
        kwargs = dict(app=self.app())
        if 'cache_service' in class_routes.__annotations__:             # if there is a Type_Safe definiton of cache_service
            kwargs["cache_service"] = self.cache_service                # we will assign it (aka dependency injection)
        class_routes(**kwargs).setup()
        return self


    # todo: refactor to separate class (focused on setting up this static route)
    def setup_web_console(self):


        path_static_folder  = mgraph_ai_service_cache__web_console.path
        path_name           = CACHE_SERVICE__WEB_CONSOLE__PATH
        path_static         = f"/{path_name}"

        major_version       = CACHE_SERVICE__WEB_CONSOLE__MAJOR__VERSION
        latest_version      = CACHE_SERVICE__WEB_CONSOLE__LATEST__VERSION
        start_page          = CACHE_SERVICE__WEB_CONSOLE__ROUTE__START_PAGE

        path_latest_version = f"/{path_name}/{major_version}/{latest_version}/{start_page}.html"
        self.app().mount(path_static, StaticFiles(directory=path_static_folder), name=path_name)


        @route_path(path=f'/{CACHE_SERVICE__WEB_CONSOLE__PATH}')
        def redirect_to_latest():
            return RedirectResponse(url=path_latest_version)

        self.add_route_get(redirect_to_latest)


# # todo: see if we need to refactor this into Fast_API since the current routes are working ok for this project (for string, bytes and dict)
# intercept the bytes submitted here
#self.app().add_middleware(BodyReaderMiddleware)

# from fastapi import Request
# from starlette.middleware.base import BaseHTTPMiddleware
# class BodyReaderMiddleware(BaseHTTPMiddleware):
#     async def dispatch(self, request: Request, call_next):
#         if request.method in ["POST", "PUT", "PATCH"]:                              # Only read body for certain methods
#             body = await request.body()
#             request.state.body = body                                               # Store it in request state for later sync access
#         else:
#             request.state.body = None
#
#         response = await call_next(request)
#         return response





import base64
import json
from typing                                                                              import Union, Dict
from fastapi                                                                             import HTTPException, Response
from memory_fs.schemas.Schema__Memory_FS__File__Metadata                                 import Schema__Memory_FS__File__Metadata
from memory_fs.schemas.Schema__Memory_FS__File__Config                                   import Schema__Memory_FS__File__Config
from osbot_utils.type_safe.Type_Safe                                                     import Type_Safe
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                           import type_safe
from mgraph_ai_service_cache_client.schemas.cache.file.Schema__Cache__File__Refs         import Schema__Cache__File__Refs
from mgraph_ai_service_cache.service.cache.Cache__Service                                import Cache__Service
from osbot_fast_api.api.decorators.route_path                                            import route_path
from osbot_fast_api.api.routes.Fast_API__Routes                                          import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix               import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag                  import Safe_Str__Fast_API__Route__Tag
from osbot_utils.decorators.methods.cache_on_self                                        import cache_on_self
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                       import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id          import Safe_Str__Id
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash import Safe_Str__Cache_Hash
from mgraph_ai_service_cache_client.schemas.cache.Schema__Cache__Binary__Reference       import Schema__Cache__Binary__Reference
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Data_Type           import Enum__Cache__Data_Type
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API                       import FAST_API__PARAM__NAMESPACE
from mgraph_ai_service_cache.service.cache.retrieve.Cache__Service__Retrieve             import Cache__Service__Retrieve

TAG__ROUTES_RETRIEVE                  = 'retrieve'
PREFIX__ROUTES_RETRIEVE               = '/{namespace}'
BASE_PATH__ROUTES_RETRIEVE            = f'{PREFIX__ROUTES_RETRIEVE}/{TAG__ROUTES_RETRIEVE}/'
ROUTES_PATHS__RETRIEVE                = [ BASE_PATH__ROUTES_RETRIEVE + '{cache_id}'                 ,
                                          BASE_PATH__ROUTES_RETRIEVE + '{cache_id}/config'          ,
                                          BASE_PATH__ROUTES_RETRIEVE + '{cache_id}/refs'            ,
                                          BASE_PATH__ROUTES_RETRIEVE + '{cache_id}/refs/all'        ,
                                          BASE_PATH__ROUTES_RETRIEVE + '{cache_id}/metadata'        ,
                                          BASE_PATH__ROUTES_RETRIEVE + '{cache_id}/binary'          ,
                                          BASE_PATH__ROUTES_RETRIEVE + '{cache_id}/json'            ,
                                          BASE_PATH__ROUTES_RETRIEVE + '{cache_id}/string'          ,
                                          BASE_PATH__ROUTES_RETRIEVE + 'hash/{cache_hash}'          ,
                                          BASE_PATH__ROUTES_RETRIEVE + 'hash/{cache_hash}/binary'   ,
                                          BASE_PATH__ROUTES_RETRIEVE + 'hash/{cache_hash}/json'     ,
                                          BASE_PATH__ROUTES_RETRIEVE + 'hash/{cache_hash}/string'   ,
                                          BASE_PATH__ROUTES_RETRIEVE + 'hash/{cache_hash}/metadata' ,
                                          BASE_PATH__ROUTES_RETRIEVE + 'hash/{cache_hash}/refs-hash',
                                          BASE_PATH__ROUTES_RETRIEVE + 'hash/{cache_hash}/cache-id']

class Routes__File__Retrieve(Fast_API__Routes):                                             # FastAPI routes for cache retrieval operations
    tag            : Safe_Str__Fast_API__Route__Tag    = TAG__ROUTES_RETRIEVE
    prefix         : Safe_Str__Fast_API__Route__Prefix = PREFIX__ROUTES_RETRIEVE
    cache_service  : Cache__Service                                                                          # get cache_service via Dependency Injection
    
    @cache_on_self
    def retrieve_service(self):                                                                             # Service layer for business logic
        return Cache__Service__Retrieve(cache_service=self.cache_service)                                   # create Cache__Service__Retrieve object (once, using the shared Cache_Service)

    @type_safe
    def handle_not_found(self, result        : Union[Type_Safe, Dict] = None,                                 # Base method for 404 handling
                               cache_id      : Cache_Id            = None,
                               cache_hash    : Safe_Str__Cache_Hash   = None,
                               namespace     : Safe_Str__Id           = None):
        if result is None:
            error = self.retrieve_service().get_not_found_error(cache_id   = cache_id  ,
                                                                cache_hash = cache_hash,
                                                                namespace  = namespace )
            raise HTTPException(status_code=404, detail=error.json())
        return result

    def retrieve__cache_id(self, cache_id  : Cache_Id,
                                 namespace : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                            ): # todo union is not supported by Fast_API. Refactor to one class -> Union[Schema__Cache__Retrieve__Success, Schema__Cache__Binary__Reference]:             # Retrieve by cache ID with metadata
                
        result = self.retrieve_service().retrieve_by_id(cache_id, namespace)                                              # Use service layer
        
        if result:
            # todo: this logic should not be happening in this routes class, this should be handled by retrieve_by_id
            if result.data_type == Enum__Cache__Data_Type.BINARY:                           # Handle binary data that can't be returned in JSON
                binary_url = f"/{namespace}/retrieve/{cache_id}/binary"
                return Schema__Cache__Binary__Reference(message      = "Binary data requires separate endpoint"    ,        # todo: refactor this to use the Schema__Cache__Retrieve__Success
                                                        data_type    = Enum__Cache__Data_Type.BINARY               ,        #       which is quite compatible with this logic
                                                        size         = result.metadata.content_size                ,        #       since this is still a retrieve success, but it just the
                                                        cache_hash   = result.metadata.cache_hash                  ,        #       edge case where we don't return a binary here
                                                        cache_id     = result.metadata.cache_id                    ,        #       (to avoid having to convert bytes into base64)
                                                        namespace    = namespace                                   ,
                                                        binary_url   = binary_url                                  ,
                                                        metadata     = result.metadata                             )
        return self.handle_not_found(result, cache_id=cache_id, namespace=namespace)

    def retrieve__cache_id__config(self, cache_id  : Cache_Id,
                                         namespace : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                    ) -> Schema__Memory_FS__File__Config:                                           # Retrieve by cache ID with metadata

        response_data = self.retrieve_service().retrieve_by_id__config(cache_id, namespace)
        return self.handle_not_found(response_data, cache_id=cache_id, namespace=namespace)

    def retrieve__cache_id__refs(self, cache_id : Cache_Id,
                                       namespace: Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                  ) -> Schema__Cache__File__Refs:                                                    # Get cache entry details

        result = self.retrieve_service().retrieve_by_id__refs(cache_id, namespace)
        return self.handle_not_found(result, cache_id=cache_id, namespace=namespace)

    def retrieve__cache_id__metadata(self, cache_id : Cache_Id,
                                           namespace: Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                      ) -> Schema__Memory_FS__File__Metadata:                                                       # Get cache entry details
        result = self.retrieve_service().retrieve_by_id__metadata(cache_id, namespace)                                              # todo: this class should return Schema__Cache__File__Refs
        return self.handle_not_found(result, cache_id=cache_id, namespace=namespace)


    def retrieve__cache_id__refs__all(self, cache_id: Cache_Id,
                                            namespace: Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                      ) -> Dict:
        result = self.retrieve_service().get_entry_details__all(cache_id, namespace)                                              # todo: this class should return Schema__Cache__File__Refs
        return self.handle_not_found(result, cache_id=cache_id, namespace=namespace)

    @route_path("/retrieve/hash/{cache_hash}")
    def retrieve__hash__cache_hash(self, cache_hash : Safe_Str__Cache_Hash,
                                         namespace  : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                    ): # todo union is not supported by Fast_API. Refactor to one class  -> Union[Schema__Cache__Retrieve__Success, Schema__Cache__Binary__Reference]:     # Retrieve by hash
                
        result = self.retrieve_service().retrieve_by_hash(cache_hash, namespace)                                          # Use service layer
        
        if result is None:
            # Return 404 Not Found
            error = self.retrieve_service().get_not_found_error(cache_hash=cache_hash, namespace=namespace)
            raise HTTPException(status_code=404, detail=error.json())
        
        # Handle binary data redirect
        if result.data_type == Enum__Cache__Data_Type.BINARY:
            binary_url = f"/{namespace}/retrieve/hash/{cache_hash}/binary"
            return Schema__Cache__Binary__Reference(
                message      = "Binary data requires separate endpoint"    ,
                data_type    = Enum__Cache__Data_Type.BINARY              ,
                size         = result.metadata.content_size               ,
                cache_hash   = result.metadata.cache_hash                 ,
                cache_id     = result.metadata.cache_id                   ,
                namespace    = namespace                                  ,
                binary_url   = binary_url                                  ,
                metadata     = result.metadata                             )
        
        return result
    
    @route_path("/retrieve/{cache_id}/string")
    def retrieve__cache_id__string(self, cache_id : Cache_Id,
                                         namespace: Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                    ) -> Response:                                                          # Retrieve as string format
        
        result = self.retrieve_service().retrieve_by_id(cache_id, namespace)
        
        if result is None:
            raise HTTPException(status_code=404, detail="Cache entry not found")
        
        # Convert data to string format
        if result.data_type == Enum__Cache__Data_Type.STRING:
            content = result.data
        elif result.data_type == Enum__Cache__Data_Type.JSON:
            content = json.dumps(result.data)
        elif result.data_type == Enum__Cache__Data_Type.BINARY:
            content = base64.b64encode(result.data).decode('utf-8')         # if it is binary encode as base64
        else:
            content = str(result.data)
        
        return Response(content=content, media_type="text/plain")
    
    @route_path("/retrieve/{cache_id}/json")
    def retrieve__cache_id__json(self, cache_id  : Cache_Id,
                                       namespace : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                  ) -> dict:                                                                    # Retrieve as JSON format
        namespace = namespace or Safe_Str__Id("default")
        
        result = self.retrieve_service().retrieve_by_id(cache_id, namespace)
        
        if result is None:
            raise HTTPException(status_code=404, detail="Cache entry not found")
        
        # Return data based on type
        if result.data_type == Enum__Cache__Data_Type.JSON:
            return result.data
        elif result.data_type == Enum__Cache__Data_Type.STRING:
            # Try to parse as JSON
            try:
                return json.loads(result.data)
            except json.JSONDecodeError:
                # Return 415 Unsupported Media Type
                raise HTTPException(status_code=415, 
                                  detail=f"Data is string, not JSON: {result.data[:100]}")
        elif result.data_type == Enum__Cache__Data_Type.BINARY:
            # Return base64 encoded in JSON wrapper
            return {
                "data_type": "binary",
                "encoding": "base64",
                "data": base64.b64encode(result.data).decode('utf-8')
            }
        
        return {"data": result.data, "data_type": str(result.data_type)}                                                    # todo: look at how to handle this scenario that shouldn't happen
    
    @route_path("/retrieve/{cache_id}/binary")
    def retrieve__cache_id__binary(self, cache_id : Cache_Id,
                                         namespace: Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                    ) -> Response:                                                                          # Retrieve as binary format
        
        result = self.retrieve_service().retrieve_by_id(cache_id, namespace)
        
        if result is None:
            raise HTTPException(status_code=404, detail="Cache entry not found")
        
        # Convert to binary format
        if result.data_type == Enum__Cache__Data_Type.BINARY:                                                               # todo: refactor to helper method since this is quite a common pattern (something like, convert_to_bytes)
            content = result.data
        elif result.data_type == Enum__Cache__Data_Type.STRING:
            content = result.data.encode('utf-8')
        elif result.data_type == Enum__Cache__Data_Type.JSON:
            content = json.dumps(result.data).encode('utf-8')
        else:
            content = str(result.data).encode('utf-8')
        
        return Response(content=content, media_type="application/octet-stream")
    
    @route_path("/retrieve/hash/{cache_hash}/string")
    def retrieve__hash__cache_hash__string(self, cache_hash : Safe_Str__Cache_Hash,
                                                 namespace  : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                            ) -> Response:                                                                  # Retrieve string by hash
        
        result = self.retrieve_service().retrieve_by_hash(cache_hash, namespace)
        
        if result is None:
            raise HTTPException(status_code=404, detail="Cache entry not found")

        if result.data_type == Enum__Cache__Data_Type.STRING:                                                               # Same conversion logic as retrieve__cache_id__string
            content = result.data
        elif result.data_type == Enum__Cache__Data_Type.JSON:                                                               # todo: refactor to common conver method
            content = json.dumps(result.data)
        elif result.data_type == Enum__Cache__Data_Type.BINARY:
            try:
                content = result.data.decode('utf-8')
            except (UnicodeDecodeError, AttributeError):
                content = base64.b64encode(result.data).decode('utf-8')
        else:
            content = str(result.data)                                                                                      # todo: look at how to handle this scenario that shouldn't happen
        
        return Response(content=content, media_type="text/plain")
    
    @route_path("/retrieve/hash/{cache_hash}/json")
    def retrieve__hash__cache_hash__json(self, cache_hash : Safe_Str__Cache_Hash,
                                               namespace  : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                          ) -> Dict:                                   # Retrieve JSON by hash

        
        result = self.retrieve_service().retrieve_by_hash(cache_hash, namespace)
        
        if result is None:
            raise HTTPException(status_code=404, detail="Cache entry not found")

        if result.data_type == Enum__Cache__Data_Type.JSON:                                                                 # Same logic as retrieve__cache_id__json
            return result.data
        elif result.data_type == Enum__Cache__Data_Type.STRING:                                                             # todo: refactor to common convert method
            try:
                return json.loads(result.data)
            except json.JSONDecodeError:
                raise HTTPException(status_code=415, 
                                    detail=f"Data is string, not JSON")
        elif result.data_type == Enum__Cache__Data_Type.BINARY:
            return { "data_type" : "binary",                                                                                # todo: convert to Type_Safe class
                     "encoding"  : "base64",
                     "data"      : base64.b64encode(result.data).decode('utf-8')  }

    @route_path("/retrieve/hash/{cache_hash}/metadata")
    def retrieve__hash__cache_hash__metadata(self, cache_hash : Safe_Str__Cache_Hash,
                                                   namespace  : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                              ) -> dict:                                   # Retrieve metadata by hash


        result = self.retrieve_service().retrieve_by_hash__metadata(cache_hash, namespace)

        if result is None:
            raise HTTPException(status_code=404, detail="Cache entry not found")
        #return result                       # todo: see why this raises the pydantic error: Input should be a valid string [type=string_type, input_value=None, input_type=NoneType]
        return result.json()                 #       this works, but it would be better to not need this (at least here)

    # todo: bug this should be just "/retrieve/hash/{cache_hash}/refs" (i.e. replace the /refs-hash with /refs)
    @route_path("/retrieve/hash/{cache_hash}/refs-hash")
    def retrieve__hash__cache_hash__refs_hash(self, cache_hash : Safe_Str__Cache_Hash,
                                                    namespace  : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                               ) -> dict:                                   # Retrieve metadata by hash


        result = self.retrieve_service().retrieve_by_hash__refs_hash(cache_hash, namespace)

        if result is None:
            raise HTTPException(status_code=404, detail="Cache entry not found")

        return result

    @route_path("/retrieve/hash/{cache_hash}/cache-id")
    def retrieve__hash__cache_hash__cache_id(self, cache_hash : Safe_Str__Cache_Hash,
                                                   namespace  : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                              ) -> Dict:                                   # Retrieve metadata by hash


        result = self.retrieve_service().retrieve_by_hash__refs_hash(cache_hash, namespace)

        if result is None:
            raise HTTPException(status_code=404, detail="Cache entry not found")

        return dict(namespace  = namespace,
                    cache_hash = cache_hash,
                    cache_id   = result.get('latest_id'))


    @route_path("/retrieve/hash/{cache_hash}/binary")
    def retrieve__hash__cache_hash__binary(self, cache_hash : Safe_Str__Cache_Hash,
                                                 namespace  : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                            ) -> Response:                             # Retrieve binary by hash
        
        result = self.retrieve_service().retrieve_by_hash(cache_hash, namespace)
        
        if result is None:
            raise HTTPException(status_code=404, detail="Cache entry not found")

        if result.data_type == Enum__Cache__Data_Type.BINARY:                                                               # Same conversion as retrieve__cache_id__binary
            content = result.data
        elif result.data_type == Enum__Cache__Data_Type.STRING:                                                             # todo: refactor to common convert method
            content = result.data.encode('utf-8')
        elif result.data_type == Enum__Cache__Data_Type.JSON:
            content = json.dumps(result.data).encode('utf-8')
        else:
            content = str(result.data).encode('utf-8')                                                                      # todo: look at how to handle this scenario that shouldn't happen
        
        return Response(content=content, media_type="application/octet-stream")
    


    def setup_routes(self):                                                                                                 # Configure all routes
        self.add_route_get(self.retrieve__cache_id                  )               # Generic retrieval (with metadata)
        self.add_route_get(self.retrieve__cache_id__config          )
        self.add_route_get(self.retrieve__cache_id__metadata        )
        self.add_route_get(self.retrieve__cache_id__refs            )
        self.add_route_get(self.retrieve__cache_id__refs__all       )

        self.add_route_get(self.retrieve__cache_id__string          )               # Type-specific retrieval
        self.add_route_get(self.retrieve__cache_id__json            )
        self.add_route_get(self.retrieve__cache_id__binary          )

        self.add_route_get(self.retrieve__hash__cache_hash           )
        self.add_route_get(self.retrieve__hash__cache_hash__string   )
        self.add_route_get(self.retrieve__hash__cache_hash__json     )
        self.add_route_get(self.retrieve__hash__cache_hash__binary   )
        self.add_route_get(self.retrieve__hash__cache_hash__metadata )
        self.add_route_get(self.retrieve__hash__cache_hash__refs_hash)
        self.add_route_get(self.retrieve__hash__cache_hash__cache_id )